@@ -28,21 +28,6 @@
 #include "gtkplotcanvas.h"
 #include "gtkplotgdk.h"
 
-#ifdef WITH_GTKGL
-#include <GL/gl.h>
-#include <GL/glu.h>
-#include <gtkgl/gdkgl.h>
-#include "gtkplotgl.h"
-
-static gint visual_attributes[] = { GDK_GL_RGBA,
-                                    GDK_GL_RED_SIZE,1,
-                                    GDK_GL_GREEN_SIZE,1,
-                                    GDK_GL_BLUE_SIZE,1,
-                                    GDK_GL_NONE };
-
-
-#endif /* WITH_GTKGL */
-
 #define DEFAULT_WIDTH 100
 #define DEFAULT_HEIGHT 150
 #define DEFAULT_MARKER_SIZE 6
@@ -89,8 +74,7 @@ static void gtk_plot_canvas_get_real_pixel      (GtkWidget *widget,
                                                  gdouble px, gdouble py,
                                                  gint *x, gint *y);
 static void draw_selection 			(GtkPlotCanvas *canvas, 
-                				 GdkRectangle area,
-                				 gboolean markers);
+                				 GdkRectangle area);
 static void draw_marker				(GtkPlotCanvas *canvas, 
 						 GdkGC *gc, gint x, gint y);
 static void gtk_plot_canvas_draw_text           (GtkPlotCanvas *canvas,
@@ -370,6 +354,7 @@ gtk_plot_canvas_init (GtkPlotCanvas *plot_canvas)
   gtk_widget_set_events (widget, gtk_widget_get_events(widget)|
                          GRAPH_MASK);
 
+  plot_canvas->freeze_count = 0;
   plot_canvas->cursor = gdk_cursor_new(GDK_TOP_LEFT_ARROW);
 
   plot_canvas->num_plots = 0;
@@ -381,7 +366,7 @@ gtk_plot_canvas_init (GtkPlotCanvas *plot_canvas)
   plot_canvas->magnification = 1.;
 
   plot_canvas->show_grid = FALSE;
-  plot_canvas->grid_step = 20;
+  plot_canvas->grid_step = 20.;
   plot_canvas->grid.line_style = GTK_PLOT_LINE_SOLID;
   plot_canvas->grid.line_width = 0;
 
@@ -406,15 +391,30 @@ gtk_plot_canvas_init (GtkPlotCanvas *plot_canvas)
 
   gtk_psfont_init();
 
-#ifdef WITH_GTKGL
-  plot_canvas->pc = GTK_PLOT_PC(gtk_plot_gl_new(NULL, visual_attributes));
-#else
-  plot_canvas->pc = GTK_PLOT_PC(gtk_plot_gdk_new(NULL));
-#endif
+  plot_canvas->pc = NULL;
+  gtk_plot_canvas_set_pc(plot_canvas, NULL);
 
   plot_canvas->pixmap = NULL;
 }
 
+void
+gtk_plot_canvas_set_pc(GtkPlotCanvas *canvas, GtkPlotPC *pc)
+{
+  if(canvas->pc)
+    gtk_object_unref(GTK_OBJECT(canvas->pc));
+
+  if(!pc){
+    canvas->pc = GTK_PLOT_PC(gtk_plot_gdk_new(NULL));
+  } else {
+    canvas->pc = pc;
+  }
+
+  if(canvas->pc && GTK_IS_PLOT_GDK(canvas->pc)){
+       GTK_PLOT_GDK(canvas->pc)->drawable = canvas->pixmap;
+  }
+  gtk_plot_pc_set_viewport(canvas->pc, canvas->pixmap_width, canvas->pixmap_height);
+}
+
 GtkWidget*
 gtk_plot_canvas_new (gint width, gint height, gdouble magnification)
 {
@@ -441,6 +441,18 @@ gtk_plot_canvas_construct(GtkPlotCanvas *plot_canvas,
   gtk_plot_canvas_set_magnification(plot_canvas, m);
 }
 
+void
+gtk_plot_canvas_freeze(GtkPlotCanvas *canvas)
+{
+  canvas->freeze_count++;
+}
+
+void
+gtk_plot_canvas_thaw(GtkPlotCanvas *canvas)
+{
+  if(canvas->freeze_count == 0) return;
+  canvas->freeze_count--;
+}
 
 static void
 gtk_plot_canvas_destroy (GtkObject *object)
@@ -486,7 +498,7 @@ gtk_plot_canvas_destroy (GtkObject *object)
   }
 
   gdk_cursor_destroy(plot_canvas->cursor);
-  gtk_object_destroy(GTK_OBJECT(plot_canvas->pc));
+  gtk_object_unref(GTK_OBJECT(plot_canvas->pc));
 
   if (GTK_OBJECT_CLASS (parent_class)->destroy)
     (*GTK_OBJECT_CLASS (parent_class)->destroy) (object);
@@ -545,12 +557,9 @@ gtk_plot_canvas_paint (GtkPlotCanvas *canvas)
   widget = GTK_WIDGET(canvas);
 
   if(GTK_WIDGET_REALIZED(widget) && !canvas->pixmap) return;
+  if(canvas->freeze_count > 0) return;
 
-#ifdef WITH_GTKGL
-  if(GTK_IS_PLOT_GL(canvas->pc)){
-    gtk_plot_gl_set_drawable(GTK_PLOT_GL(canvas->pc), canvas->pixmap);
-  }
-#endif
+  gtk_plot_pc_init(canvas->pc);
 
   gtk_plot_pc_gsave(canvas->pc);
   gtk_plot_pc_set_color(canvas->pc, &canvas->background);
@@ -600,6 +609,8 @@ gtk_plot_canvas_refresh(GtkPlotCanvas *canvas)
 {
   if(!GTK_WIDGET_REALIZED(GTK_WIDGET(canvas))) return;
   if(!canvas->pixmap) return;
+  if(canvas->freeze_count > 0) return;
+
   gdk_draw_pixmap(GTK_WIDGET(canvas)->window,
                   GTK_WIDGET(canvas)->style->fg_gc[GTK_STATE_NORMAL],
                   canvas->pixmap,
@@ -611,7 +622,7 @@ gtk_plot_canvas_refresh(GtkPlotCanvas *canvas)
 static void
 gtk_plot_canvas_draw_grid(GtkPlotCanvas *canvas)
 {
-  gint x, y;
+  gdouble x, y;
 
   if(!canvas->pixmap) return;
   if(!canvas->show_grid) return;
@@ -621,10 +632,12 @@ gtk_plot_canvas_draw_grid(GtkPlotCanvas *canvas)
   gtk_plot_canvas_set_line_attributes(canvas, canvas->grid);
 
   for(x = 0; x < canvas->pixmap_width; x += canvas->grid_step)
-      gtk_plot_pc_draw_line(canvas->pc, x, 0, x, canvas->pixmap_height);
+      gtk_plot_pc_draw_line(canvas->pc, 
+                            roundint(x), 0, roundint(x), canvas->pixmap_height);
 
   for(y = 0; y < canvas->pixmap_height; y += canvas->grid_step)
-      gtk_plot_pc_draw_line(canvas->pc, 0, y, canvas->pixmap_width, y);
+      gtk_plot_pc_draw_line(canvas->pc, 
+                            0, roundint(y), canvas->pixmap_width, roundint(y));
 }
 
 static void
@@ -660,8 +673,12 @@ gtk_plot_canvas_motion (GtkWidget *widget, GdkEventMotion *event)
   gint x, y;
   gint new_x, new_y;
   gint new_width, new_height;
-  gint cursor;
+  gint cursor = GDK_TOP_LEFT_ARROW;
   gint pivot_x, pivot_y;
+  gdouble fx, fy, fz, fa;
+  gdouble fdx, fdy, fdz, fda;
+  gchar *label;
+  gboolean error;
 
   canvas = GTK_PLOT_CANVAS(widget);
   gtk_widget_get_pointer(widget, &x, &y);
@@ -676,6 +693,8 @@ gtk_plot_canvas_motion (GtkWidget *widget, GdkEventMotion *event)
 
   if(canvas->action == GTK_PLOT_CANVAS_ACTION_INACTIVE) 
        cursor = GDK_TOP_LEFT_ARROW;
+  else if(canvas->action == GTK_PLOT_CANVAS_ACTION_DRAG) 
+       cursor = GDK_FLEUR;
   else
        switch(canvas->drag_point){
             case GTK_PLOT_CANVAS_TOP_LEFT: 
@@ -741,14 +760,73 @@ gtk_plot_canvas_motion (GtkWidget *widget, GdkEventMotion *event)
        if(canvas->active_item.flags & GTK_PLOT_CANVAS_CAN_MOVE){
          gint dx, dy;
 
-         draw_selection(canvas, canvas->drag_area, TRUE);
+         if(canvas->active_item.type == GTK_PLOT_CANVAS_MARKER){
+           GtkPlotMarker *marker;
+           gboolean move = FALSE;
+           gint i;
+
+           marker = (GtkPlotMarker *)canvas->active_item.data;
+           for(i = 0; i < marker->data->num_points; i++){
+              gdouble px, py;
+	      gtk_plot_data_get_point(marker->data, i,
+	  			     &fx, &fy, &fz, &fa, &fdx, &fdy, &fdz, &fda,
+				     &label, &error); 
+              gtk_plot_get_pixel(marker->data->plot,
+                                 fx, fy,
+                                 &px, &py);
+             
+              if(fabs(x - px) <= DEFAULT_MARKER_SIZE && 
+                 fabs(y - py) <= DEFAULT_MARKER_SIZE) {
+                      move = TRUE;
+                      x = px;
+                      y = py;
+                      break;
+              }
+           }
+           if(!move) return TRUE;
+
+         } else if(canvas->active_item.type == GTK_PLOT_CANVAS_DATA){
+	   gboolean veto = FALSE;
+
+           draw_selection(canvas, canvas->drag_area);
+		
+	   gtk_plot_get_point(canvas->active_plot, x, y, &fx, &fy);
+
+           gtk_signal_emit(GTK_OBJECT(canvas), canvas_signals[MOVE_ITEM],
+                           &canvas->active_item, fx, fy, &veto);
+
+           if(veto){
+              gtk_signal_emit (GTK_OBJECT(canvas), canvas_signals[CHANGED]);
+
+              gtk_plot_data_get_point(canvas->active_data, canvas->active_point,
+	    		             &fx, &fy, &fz, &fa, &fdx, &fdy, &fdz, &fda,
+			             &label, &error);
+              if(fx != canvas->active_x || fy != canvas->active_y){
+                canvas->active_x = fx;
+                canvas->active_y = fy;
+
+	        canvas->drag_x = x;
+	        canvas->drag_y = y;
+
+	        canvas->drag_area.x = x - 10;
+	        canvas->drag_area.y = y - 10;
+
+                canvas->active_item.allocation.x = x - 10;
+                canvas->active_item.allocation.y = y - 10;
+                canvas->active_item.allocation.width = 20;
+             }
+           }
+           draw_selection(canvas, canvas->drag_area);
+           return TRUE;
+         }
+         draw_selection(canvas, canvas->drag_area);
          canvas->pointer_x = x;
          canvas->pointer_y = y;
          dx = x - canvas->drag_x;
          dy = y - canvas->drag_y;
          area.x = canvas->active_item.allocation.x + dx; 
          area.y = canvas->active_item.allocation.y + dy;
-         draw_selection(canvas, area, TRUE);
+         draw_selection(canvas, area);
          canvas->drag_area = area;
        }
        break;
@@ -800,19 +878,19 @@ gtk_plot_canvas_motion (GtkWidget *widget, GdkEventMotion *event)
 
                 canvas->pointer_x = x;
                 canvas->pointer_y = y;
-                draw_selection(canvas, canvas->drag_area, TRUE);
+                draw_selection(canvas, canvas->drag_area);
                 area.x = new_x;
                 area.y = new_y;
                 area.width = new_width;
                 area.height = new_height;
-                draw_selection(canvas, area, TRUE);
+                draw_selection(canvas, area);
                 canvas->drag_area = area;
 
        }
 
        break;
      case GTK_PLOT_CANVAS_ACTION_SELECTION:
-         draw_selection(canvas, canvas->drag_area, FALSE);
+         draw_selection(canvas, canvas->drag_area);
          canvas->pointer_x = x;
          canvas->pointer_y = y;
          area.x = MIN(canvas->pointer_x, canvas->drag_x); 
@@ -820,7 +898,7 @@ gtk_plot_canvas_motion (GtkWidget *widget, GdkEventMotion *event)
          area.width = abs(x - canvas->drag_x);
          area.height = abs(y - canvas->drag_y);
          canvas->drag_area = area;
-         draw_selection(canvas, canvas->drag_area, FALSE);
+         draw_selection(canvas, canvas->drag_area);
  	 break;
      case GTK_PLOT_CANVAS_ACTION_INACTIVE:
      default:
@@ -914,6 +992,8 @@ gtk_plot_canvas_button_press(GtkWidget *widget, GdkEventButton *event)
           active_item.type = child->type;
           active_item.state = GTK_STATE_SELECTED;
           active_item.flags = child->flags;
+          active_item.selection = child->selection;
+          active_item.mode = child->mode;
           state = GTK_STATE_SELECTED;
           new_selection = TRUE;
           new_pos = pos;
@@ -966,6 +1046,8 @@ gtk_plot_canvas_button_press(GtkWidget *widget, GdkEventButton *event)
                 active_item.type = GTK_PLOT_CANVAS_TITLE;
                 active_item.state = GTK_STATE_SELECTED;
                 active_item.flags = GTK_PLOT_CANVAS_CAN_MOVE;
+                active_item.selection = GTK_PLOT_CANVAS_SELECT_MARKERS;
+                active_item.mode = GTK_PLOT_CANVAS_SELECT_CLICK_2;
                 state = GTK_STATE_SELECTED;
                 click_plot = plot;
                 new_selection = TRUE;
@@ -1000,6 +1082,8 @@ gtk_plot_canvas_button_press(GtkWidget *widget, GdkEventButton *event)
            active_item.type = GTK_PLOT_CANVAS_LEGENDS;
            active_item.state = GTK_STATE_SELECTED;
            active_item.flags = GTK_PLOT_CANVAS_CAN_MOVE;
+           active_item.selection = GTK_PLOT_CANVAS_SELECT_MARKERS;
+           active_item.mode = GTK_PLOT_CANVAS_SELECT_CLICK_2;
            state = GTK_STATE_SELECTED;
            click_plot = plot;
            new_selection = TRUE;
@@ -1035,6 +1119,8 @@ gtk_plot_canvas_button_press(GtkWidget *widget, GdkEventButton *event)
                 active_item.type = GTK_PLOT_CANVAS_AXIS;
                 active_item.state = GTK_STATE_SELECTED;
                 active_item.flags = GTK_PLOT_CANVAS_FROZEN;
+                active_item.selection = GTK_PLOT_CANVAS_SELECT_MARKERS;
+                active_item.mode = GTK_PLOT_CANVAS_SELECT_CLICK_2;
                 state = GTK_STATE_SELECTED;
                 new_pos = pos;
                 click_plot = plot;
@@ -1046,6 +1132,8 @@ gtk_plot_canvas_button_press(GtkWidget *widget, GdkEventButton *event)
                 active_item.type = GTK_PLOT_CANVAS_AXIS;
                 active_item.state = GTK_STATE_SELECTED;
                 active_item.flags = GTK_PLOT_CANVAS_FROZEN;
+                active_item.selection = GTK_PLOT_CANVAS_SELECT_MARKERS;
+                active_item.mode = GTK_PLOT_CANVAS_SELECT_CLICK_2;
                 state = GTK_STATE_SELECTED;
                 new_pos = pos;
                 click_plot = plot;
@@ -1060,6 +1148,8 @@ gtk_plot_canvas_button_press(GtkWidget *widget, GdkEventButton *event)
                 active_item.type = GTK_PLOT_CANVAS_AXIS;
                 active_item.state = GTK_STATE_SELECTED;
                 active_item.flags = GTK_PLOT_CANVAS_FROZEN;
+                active_item.selection = GTK_PLOT_CANVAS_SELECT_MARKERS;
+                active_item.mode = GTK_PLOT_CANVAS_SELECT_CLICK_2;
                 state = GTK_STATE_SELECTED;
                 new_pos = pos;
                 click_plot = plot;
@@ -1071,6 +1161,8 @@ gtk_plot_canvas_button_press(GtkWidget *widget, GdkEventButton *event)
                 active_item.type = GTK_PLOT_CANVAS_AXIS;
                 active_item.state = GTK_STATE_SELECTED;
                 active_item.flags = GTK_PLOT_CANVAS_FROZEN;
+                active_item.selection = GTK_PLOT_CANVAS_SELECT_MARKERS;
+                active_item.mode = GTK_PLOT_CANVAS_SELECT_CLICK_2;
                 state = GTK_STATE_SELECTED;
                 new_pos = pos;
                 click_plot = plot;
@@ -1092,6 +1184,8 @@ gtk_plot_canvas_button_press(GtkWidget *widget, GdkEventButton *event)
               active_item.flags = GTK_PLOT_CANVAS_CAN_MOVE |
                                   GTK_PLOT_CANVAS_CAN_X_RESIZE |
                                   GTK_PLOT_CANVAS_CAN_Y_RESIZE;
+              active_item.selection = GTK_PLOT_CANVAS_SELECT_MARKERS;
+              active_item.mode = GTK_PLOT_CANVAS_SELECT_CLICK_2;
               state = GTK_STATE_SELECTED;
               new_pos = pos;
               click_plot = plot;
@@ -1104,16 +1198,75 @@ gtk_plot_canvas_button_press(GtkWidget *widget, GdkEventButton *event)
 /**********************************************************************/
   if(!active_plot) active_plot = click_plot;
 
+/* DATA MARKERS */
+
   if(active_plot && !GTK_IS_PLOT3D(active_plot)){
     dataset = active_plot->data_sets;
+    while(dataset){
+      GList *markers;
+
+      data = GTK_PLOT_DATA(dataset->data);
+  
+      markers = data->markers;
+      while(markers){
+        GtkPlotMarker *marker;
+	gdouble fx, fy, fz, fa, fdx, fdy, fdz, fda;
+	gboolean error;
+	gchar *label;
+  
+        marker = (GtkPlotMarker *)markers->data;
+      
+	gtk_plot_data_get_point(data, marker->point,
+				&fx, &fy, &fz, &fa, &fdx, &fdy, &fdz, &fda,
+				&label, &error); 
+        gtk_plot_get_pixel(GTK_PLOT(active_plot), 
+                           fx, fy, &xi, &yi);
+
+        if(abs(xi-x) <= 20 && abs(yi-y) <= 20){
+  
+           active_item.type = GTK_PLOT_CANVAS_MARKER;
+           active_item.state = GTK_STATE_SELECTED;
+           active_item.flags = GTK_PLOT_CANVAS_CAN_MOVE;
+           x = xi;
+           y = yi;
+           active_item.allocation.x = xi - 10;
+           active_item.allocation.y = yi - 10;
+           active_item.allocation.width = 20;
+           active_item.allocation.height = 20;
+           active_item.data = marker;
+           active_item.selection = GTK_PLOT_CANVAS_SELECT_TARGET;
+           active_item.mode = GTK_PLOT_CANVAS_SELECT_CLICK_1;
+           new_pos = GTK_PLOT_CANVAS_IN;
+           state = GTK_STATE_SELECTED;
+           break;
+        }
+        markers = markers->next;
+      }
+      if(active_item.type == GTK_PLOT_CANVAS_MARKER) break;
+      dataset = dataset->next;
+    }
+  }
+
+/* DATA POINTS */
+
+  if(active_plot && !GTK_IS_PLOT3D(active_plot) &&
+     GTK_PLOT_CANVAS_CAN_SELECT_POINT(canvas)){
+    dataset = active_plot->data_sets;
 
     while(dataset)
      {
        data = GTK_PLOT_DATA(dataset->data);
 
        if(!data->is_function){
+	 gdouble fx, fy, fz, fa, fdx, fdy, fdz, fda;
+	 gboolean error;
+	 gchar *label;
+
          for(i = 0; i < data->num_points; i++){
-           gtk_plot_get_pixel(GTK_PLOT(active_plot), data->x[i], data->y[i], &xi, &yi);
+	   gtk_plot_data_get_point(data, i,
+	  			   &fx, &fy, &fz, &fa, &fdx, &fdy, &fdz, &fda,
+				   &label, &error); 
+           gtk_plot_get_pixel(GTK_PLOT(active_plot), fx, fy, &xi, &yi);
            if(abs(xi-x) <= DEFAULT_MARKER_SIZE && 
               abs(yi-y) <= DEFAULT_MARKER_SIZE){
                 gdouble ax, ay;
@@ -1126,11 +1279,13 @@ gtk_plot_canvas_button_press(GtkWidget *widget, GdkEventButton *event)
                                    &ax, &ay);
                 x = ax;
                 y = ay;
-                active_item.allocation.x = xi;
-                active_item.allocation.y = yi;
-                active_item.allocation.width = 20;
-                active_item.allocation.height = 20;
+                active_item.allocation.x = xi - DEFAULT_MARKER_SIZE;
+                active_item.allocation.y = yi - DEFAULT_MARKER_SIZE;
+                active_item.allocation.width = 2 * DEFAULT_MARKER_SIZE;
+                active_item.allocation.height = 2 * DEFAULT_MARKER_SIZE;
                 active_item.data = data;
+                active_item.selection = GTK_PLOT_CANVAS_SELECT_TARGET;
+                active_item.mode = GTK_PLOT_CANVAS_SELECT_CLICK_2;
                 new_pos = GTK_PLOT_CANVAS_IN;
                 state = GTK_STATE_SELECTED;
                 break;
@@ -1155,10 +1310,18 @@ gtk_plot_canvas_button_press(GtkWidget *widget, GdkEventButton *event)
 
     if(active_item.type == GTK_PLOT_CANVAS_DATA)
        {
+	        gdouble fx, fy, fz, fa, fdx, fdy, fdz, fda;
+   	        gboolean error;
+	        gchar *label;
+
+	        gtk_plot_data_get_point(data, i,
+	  	  		   &fx, &fy, &fz, &fa, &fdx, &fdy, &fdz, &fda,
+				   &label, &error); 
+
                 canvas->active_data = data;
                 canvas->active_point = i;
-                canvas->active_x = data->x[i];
-                canvas->active_y = data->y[i];
+                canvas->active_x = fx;
+                canvas->active_y = fy;
        }
 
     if(click_plot) canvas->active_plot = click_plot;
@@ -1180,6 +1343,7 @@ gtk_plot_canvas_button_press(GtkWidget *widget, GdkEventButton *event)
        active_item.type != GTK_PLOT_CANVAS_LINE &&
        active_item.type != GTK_PLOT_CANVAS_CUSTOM &&
        active_item.type != GTK_PLOT_CANVAS_PIXMAP &&
+       active_item.type != GTK_PLOT_CANVAS_MARKER &&
        click_plot != canvas->active_plot) new_item = TRUE;
 
     if(canvas->active_item.type == GTK_PLOT_CANVAS_DATA &&
@@ -1202,9 +1366,10 @@ gtk_plot_canvas_button_press(GtkWidget *widget, GdkEventButton *event)
            canvas->state = GTK_STATE_SELECTED; 
            canvas->action = GTK_PLOT_CANVAS_ACTION_INACTIVE;
 
-           draw_selection(canvas, active_item.allocation, TRUE);
+           draw_selection(canvas, active_item.allocation);
 
-           return TRUE;
+           if(active_item.mode == GTK_PLOT_CANVAS_SELECT_CLICK_2)
+             return TRUE;
 
          } else {
 
@@ -1218,18 +1383,24 @@ gtk_plot_canvas_button_press(GtkWidget *widget, GdkEventButton *event)
          }
     }
 
-    if(!new_item &&
-       ((active_item.type == GTK_PLOT_CANVAS_DATA &&
-        GTK_PLOT_CANVAS_CAN_DND_POINT(canvas)) ||
-        (active_item.type != GTK_PLOT_CANVAS_DATA &&
-         GTK_PLOT_CANVAS_CAN_DND(canvas)))) {
+    if((!new_item && active_item.mode == GTK_PLOT_CANVAS_SELECT_CLICK_2) ||   
+       active_item.mode == GTK_PLOT_CANVAS_SELECT_CLICK_1) {
+       
+      if((active_item.type == GTK_PLOT_CANVAS_DATA &&
+          GTK_PLOT_CANVAS_CAN_DND_POINT(canvas)) ||
+         (active_item.type != GTK_PLOT_CANVAS_DATA &&
+          GTK_PLOT_CANVAS_CAN_DND(canvas))) {
 
          switch(new_pos){
           case GTK_PLOT_CANVAS_IN:
               canvas->action = GTK_PLOT_CANVAS_ACTION_DRAG;
               break;
           default:
-              canvas->action = GTK_PLOT_CANVAS_ACTION_RESIZE;
+              if(active_item.flags & GTK_PLOT_CANVAS_CAN_X_RESIZE ||
+                 active_item.flags & GTK_PLOT_CANVAS_CAN_Y_RESIZE)
+                 canvas->action = GTK_PLOT_CANVAS_ACTION_RESIZE;
+              else
+                 canvas->action = GTK_PLOT_CANVAS_ACTION_DRAG;
          }
          gdk_pointer_grab (widget->window, FALSE,
                            GDK_POINTER_MOTION_HINT_MASK |
@@ -1237,14 +1408,15 @@ gtk_plot_canvas_button_press(GtkWidget *widget, GdkEventButton *event)
                            GDK_BUTTON_RELEASE_MASK,
                            NULL, NULL, event->time);
          canvas->drag_point = new_pos;
+
          canvas->drag_x = x;
          canvas->drag_y = y;
          canvas->pointer_x = x;
          canvas->pointer_y = y;
 
          return TRUE;
+      }
     }
-
   } 
 
   gtk_plot_canvas_unselect(canvas);
@@ -1260,6 +1432,7 @@ gtk_plot_canvas_button_press(GtkWidget *widget, GdkEventButton *event)
          active_item.state = GTK_STATE_SELECTED;
          active_item.flags = GTK_PLOT_CANVAS_CAN_X_RESIZE |
                              GTK_PLOT_CANVAS_CAN_Y_RESIZE;
+         active_item.selection = GTK_PLOT_CANVAS_SELECT_NONE;
          veto = TRUE;
 
          gtk_signal_emit(GTK_OBJECT(canvas), canvas_signals[SELECT_ITEM],
@@ -1281,7 +1454,7 @@ gtk_plot_canvas_button_press(GtkWidget *widget, GdkEventButton *event)
                              GDK_BUTTON1_MOTION_MASK |
                              GDK_BUTTON_RELEASE_MASK,
                              NULL, NULL, event->time);
-           draw_selection(canvas, active_item.allocation, FALSE);
+           draw_selection(canvas, active_item.allocation);
          }
   }
 
@@ -1294,11 +1467,16 @@ gtk_plot_canvas_button_release(GtkWidget *widget, GdkEventButton *event)
 {
   GtkPlotCanvas *canvas;
   GtkPlotCanvasChild *child;
+  GtkPlotMarker *marker = NULL;
   GtkPlotText *text;
   GtkPlotAxis *axis = NULL;
   GtkPlot *active_plot;
-  gdouble x, y;
+  gint active_point = 0;
+  gdouble fx, fy, fz, fa, fdx, fdy, fdz, fda;
   gdouble dx, dy;
+  gdouble x, y;
+  gboolean error;
+  gchar *label;
   gdouble new_x, new_y;
   gdouble new_width, new_height;
   gdouble x1 = 0., y1 = 0., x2 = 0., y2 = 0.;
@@ -1357,6 +1535,46 @@ gtk_plot_canvas_button_release(GtkWidget *widget, GdkEventButton *event)
 
          gtk_signal_emit (GTK_OBJECT(canvas), canvas_signals[CHANGED]);
 
+         break;
+    case GTK_PLOT_CANVAS_MARKER:
+         marker = (GtkPlotMarker *)canvas->active_item.data;
+         active_point = marker->point;
+         {
+           gboolean move = FALSE;
+           gint i;
+
+           x = canvas->pointer_x; 
+           y = canvas->pointer_y; 
+           for(i = 0; i < marker->data->num_points; i++){
+              gdouble px, py;
+
+	      gtk_plot_data_get_point(marker->data, i,
+	  	  		   &fx, &fy, &fz, &fa, &fdx, &fdy, &fdz, &fda,
+				   &label, &error); 
+              gtk_plot_get_pixel(marker->data->plot,
+                                 fx, fy, &px, &py);
+             
+              if(fabs(x - px) <= DEFAULT_MARKER_SIZE && 
+                 fabs(y - py) <= DEFAULT_MARKER_SIZE) {
+                      move = TRUE;
+                      x = px;
+                      y = py;
+                      break;
+              }
+           }
+           if(!move) break;
+           active_point = i;
+         }
+         if(active_point == marker->point) break;
+
+	 gtk_plot_data_get_point(marker->data, active_point,
+	 	  		 &fx, &fy, &fz, &fa, &fdx, &fdy, &fdz, &fda,
+				 &label, &error); 
+         gtk_signal_emit(GTK_OBJECT(canvas), canvas_signals[MOVE_ITEM],
+                         &canvas->active_item, fx, fy, &veto);
+         if(!veto) break;
+         marker->point = active_point;
+         gtk_signal_emit (GTK_OBJECT(canvas), canvas_signals[CHANGED]);
          break;
     case GTK_PLOT_CANVAS_RECTANGLE:
     case GTK_PLOT_CANVAS_LINE:
@@ -1374,7 +1592,7 @@ gtk_plot_canvas_button_release(GtkWidget *widget, GdkEventButton *event)
          y2 = y1 + new_height;
          gtk_signal_emit(GTK_OBJECT(canvas), canvas_signals[MOVE_ITEM],
                          child, 
-                         x1, x2, &veto);
+                         x2, y2, &veto);
          if(canvas->action != GTK_PLOT_CANVAS_ACTION_DRAG)
             gtk_signal_emit(GTK_OBJECT(canvas), 
                             canvas_signals[RESIZE_ITEM],
@@ -1432,29 +1650,19 @@ gtk_plot_canvas_button_release(GtkWidget *widget, GdkEventButton *event)
                               new_y,
                               new_width,
                               new_height);
+
          gtk_signal_emit (GTK_OBJECT(canvas), canvas_signals[CHANGED]);
          break;
     case GTK_PLOT_CANVAS_DATA:
          if(canvas->active_data == NULL) break;
          if(canvas->active_point == -1) break;
 
-         gtk_plot_get_pixel(GTK_PLOT(canvas->active_plot), 
-                            canvas->active_data->x[canvas->active_point],
-                            canvas->active_data->y[canvas->active_point],
-                            &x, &y);
-
-         x += (canvas->pointer_x - canvas->drag_x);
-         y += (canvas->pointer_y - canvas->drag_y);
-
-         gtk_plot_get_point(canvas->active_plot, x, y, &new_x, &new_y);
+         gtk_plot_get_point(canvas->active_plot, canvas->drag_x, canvas->drag_y, &new_x, &new_y);
 
          gtk_signal_emit(GTK_OBJECT(canvas), canvas_signals[MOVE_ITEM],
                          &canvas->active_item, new_x, new_y, &veto);
          if(!veto) break;
 
-         canvas->active_data->x[canvas->active_point] = new_x;
-         canvas->active_data->y[canvas->active_point] = new_y;
-
          gtk_signal_emit (GTK_OBJECT(canvas), canvas_signals[CHANGED]);
 
          break;
@@ -1492,7 +1700,7 @@ gtk_plot_canvas_button_release(GtkWidget *widget, GdkEventButton *event)
           canvas->action == GTK_PLOT_CANVAS_ACTION_SELECTION) ||
          (canvas->active_item.type != GTK_PLOT_CANVAS_DATA &&
           canvas->action != GTK_PLOT_CANVAS_ACTION_SELECTION))
-             draw_selection(canvas, canvas->active_item.allocation, TRUE);
+             draw_selection(canvas, canvas->active_item.allocation);
   } else {
       canvas->state = GTK_STATE_NORMAL;
   }  
@@ -1632,10 +1840,7 @@ gtk_plot_canvas_unselect (GtkPlotCanvas *plot_canvas)
 {
 
   if(plot_canvas->state == GTK_STATE_SELECTED){
-    if(plot_canvas->active_item.type == GTK_PLOT_CANVAS_NONE)
-       draw_selection(plot_canvas, plot_canvas->drag_area, FALSE); 
-    else
-       draw_selection(plot_canvas, plot_canvas->drag_area, TRUE); 
+    draw_selection(plot_canvas, plot_canvas->drag_area); 
   }
 
   plot_canvas->action = GTK_PLOT_CANVAS_ACTION_INACTIVE;
@@ -1689,7 +1894,7 @@ gtk_plot_canvas_get_active_point(GtkPlotCanvas *canvas, gdouble *x, gdouble *y)
 GtkPlotCanvasChild *
 gtk_plot_canvas_get_active_item(GtkPlotCanvas *canvas)
 {
-  return &canvas->active_item;
+  return (&canvas->active_item);
 }
 
 void
@@ -1719,6 +1924,7 @@ gtk_plot_canvas_set_size(GtkPlotCanvas *canvas, gint width, gint height)
  
        gtk_widget_set_usize(GTK_WIDGET(plot), 
                             canvas->pixmap_width, canvas->pixmap_height); 
+       gtk_signal_emit_by_name(GTK_OBJECT(plot), "update");
 
        plots = plots->next;
      }
@@ -1751,6 +1957,7 @@ gtk_plot_canvas_create_pixmap(GtkWidget *widget, gint width, gint height)
   if(canvas->pc && GTK_IS_PLOT_GDK(canvas->pc)){
        GTK_PLOT_GDK(canvas->pc)->drawable = canvas->pixmap;
   }
+  gtk_plot_pc_set_viewport(canvas->pc, width, height);
 
   gtk_plot_pc_set_color(canvas->pc, &canvas->background);
 
@@ -1838,10 +2045,9 @@ gtk_plot_canvas_set_magnification(GtkPlotCanvas *canvas,
 
 static void
 draw_selection (GtkPlotCanvas *canvas, 
-                GdkRectangle area,
-                gboolean markers)
+                GdkRectangle area)
 {
-  GdkGC *xor_gc, *gc;
+  GdkGC *xor_gc = NULL, *gc = NULL;
   GdkGCValues values;
 
   gdk_gc_get_values(GTK_WIDGET(canvas)->style->fg_gc[0], &values);
@@ -1854,13 +2060,13 @@ draw_selection (GtkPlotCanvas *canvas,
                                   GDK_GC_FUNCTION |
                                   GDK_GC_SUBWINDOW);
 
-  if(canvas->active_item.type == GTK_PLOT_CANVAS_DATA){
-    gint x, y;
-    gdouble old_x, old_y;
-    gdouble new_x, new_y;
+  if(canvas->active_item.type == GTK_PLOT_CANVAS_DATA ||
+     canvas->active_item.selection == GTK_PLOT_CANVAS_SELECT_TARGET){
 
-    area.x -= area.width / 2;
-    area.y -= area.height / 2;
+    area.x += area.width / 2 - 10;
+    area.y += area.height / 2 - 10;
+    area.width = 20;
+    area.height = 20;
     gdk_draw_rectangle (GTK_WIDGET(canvas)->window,
                         xor_gc,
                         FALSE,
@@ -1880,40 +2086,27 @@ draw_selection (GtkPlotCanvas *canvas,
                   area.x + area.width/2, area.y + area.height - 1, 
                   area.x + area.width/2, area.y + area.height - 6);
 
-    if(canvas->action == GTK_PLOT_CANVAS_ACTION_DRAG){
-      gdouble ax, ay;
-
-      old_x = canvas->active_data->x[canvas->active_point];
-      old_y = canvas->active_data->y[canvas->active_point];
-
-      gtk_plot_get_pixel(GTK_PLOT(canvas->active_plot), old_x, old_y, &ax, &ay);
-
-      x = ax + canvas->pointer_x - canvas->drag_x;
-      y = ay + canvas->pointer_y - canvas->drag_y;
-
-      gtk_plot_get_point(canvas->active_plot, x, y, &new_x, &new_y);
-
-      canvas->active_data->x[canvas->active_point] = new_x;
-      canvas->active_data->y[canvas->active_point] = new_y;
+    if(canvas->active_item.type == GTK_PLOT_CANVAS_DATA &&
+       canvas->action == GTK_PLOT_CANVAS_ACTION_DRAG){
 
       gtk_plot_set_drawable(canvas->active_plot, GTK_WIDGET(canvas)->window);
      
       gc = GTK_PLOT_GDK(canvas->active_data->plot->pc)->gc;
+      gdk_gc_ref(xor_gc);
       GTK_PLOT_GDK(canvas->active_data->plot->pc)->gc = xor_gc;
  
       gtk_plot_data_draw_points(canvas->active_data, canvas->active_data->num_points);
+
       GTK_PLOT_GDK(canvas->active_data->plot->pc)->gc = gc;
       gtk_plot_set_drawable(canvas->active_plot, canvas->pixmap);
-
-      canvas->active_data->x[canvas->active_point] = old_x;
-      canvas->active_data->y[canvas->active_point] = old_y;
     }
 
+    if(xor_gc) gdk_gc_destroy(xor_gc);
     return;
   }
 
 
-  if(markers){
+  if(canvas->active_item.selection == GTK_PLOT_CANVAS_SELECT_MARKERS){
      gdk_draw_rectangle (GTK_WIDGET(canvas)->window,
                          xor_gc,
                          FALSE,
@@ -1942,7 +2135,7 @@ draw_selection (GtkPlotCanvas *canvas,
                          area.x, area.y,
                          area.width, area.height);         
   }
-  gdk_gc_unref(xor_gc);
+  if(xor_gc) gdk_gc_destroy(xor_gc);
 }
 
 static void
@@ -2020,7 +2213,7 @@ gtk_plot_canvas_grid_set_visible(GtkPlotCanvas *canvas, gboolean visible)
 }
 
 void
-gtk_plot_canvas_grid_set_step(GtkPlotCanvas *canvas, gint step)
+gtk_plot_canvas_grid_set_step(GtkPlotCanvas *canvas, gdouble step)
 {
   canvas->grid_step = step;
   gtk_signal_emit (GTK_OBJECT(canvas), canvas_signals[CHANGED]);
@@ -2283,6 +2476,9 @@ gtk_plot_canvas_child_new(GtkPlotCanvasType type)
   child->min_width = -1;
   child->min_height = -1;
 
+  child->selection = GTK_PLOT_CANVAS_SELECT_MARKERS;
+  child->mode = GTK_PLOT_CANVAS_SELECT_CLICK_2;
+
   switch(type){
     case GTK_PLOT_CANVAS_LINE:
 	child->data = g_new0(GtkPlotCanvasLine, 1);
@@ -2412,6 +2608,21 @@ gtk_plot_canvas_remove_child(GtkPlotCanvas *canvas,
 
 }
 
+void
+gtk_plot_canvas_child_set_selection (GtkPlotCanvasChild *child,
+				     GtkPlotCanvasSelection selection)
+{
+  if(!child) return;
+  child->selection = selection;
+}
+
+void
+gtk_plot_canvas_child_set_selection_mode (GtkPlotCanvasChild *child,
+				          GtkPlotCanvasSelectionMode mode)
+{
+  if(!child) return;
+  child->mode = mode;
+}
 
 /**********************************************************************/
 static void
@@ -2423,6 +2634,8 @@ gtk_plot_canvas_draw_child(GtkPlotCanvas *canvas,
   GtkPlotCanvasEllipse *ellipse = NULL;
   GtkPlotText *text = NULL;
   GtkPlotPoint arrow[3];
+  GdkColormap *colormap = NULL;
+  GdkColor white, black;
   gint rx1 = 0, ry1 = 0, rx2 = 0, ry2 = 0;
   gint xmin, xmax, ymin, ymax;
   gint width = 0, height = 0;
@@ -2469,10 +2682,10 @@ gtk_plot_canvas_draw_child(GtkPlotCanvas *canvas,
            arrow[1].y = ry2;
            xm = rx2 - cos(angle) * line->arrow_length * m;
            ym = ry2 - sin(angle) * line->arrow_length * m;
-           arrow[0].x = xm + sin(angle)* line->arrow_width * m / 2.0;
-           arrow[0].y = ym - cos(angle)* line->arrow_width * m / 2.0;
-           arrow[2].x = xm - sin(angle)* line->arrow_width * m / 2.0;
-           arrow[2].y = ym + cos(angle)* line->arrow_width * m / 2.0;
+           arrow[0].x = xm - sin(angle)* line->arrow_width * m / 2.0;
+           arrow[0].y = ym + cos(angle)* line->arrow_width * m / 2.0;
+           arrow[2].x = xm + sin(angle)* line->arrow_width * m / 2.0;
+           arrow[2].y = ym - cos(angle)* line->arrow_width * m / 2.0;
            switch(line->arrow_style){
              case GTK_PLOT_SYMBOL_EMPTY:
                gtk_plot_pc_draw_lines (canvas->pc, arrow, 3);
@@ -2492,10 +2705,10 @@ gtk_plot_canvas_draw_child(GtkPlotCanvas *canvas,
            arrow[1].y = ry1;
            xm = rx1 + cos(angle) * line->arrow_length * m;
            ym = ry1 + sin(angle) * line->arrow_length * m;
-           arrow[0].x = xm + sin(angle)* line->arrow_width * m;
-           arrow[0].y = ym - cos(angle)* line->arrow_width * m;
-           arrow[2].x = xm - sin(angle)* line->arrow_width * m;
-           arrow[2].y = ym + cos(angle)* line->arrow_width * m;
+           arrow[0].x = xm + sin(angle)* line->arrow_width * m / 2.0;
+           arrow[0].y = ym - cos(angle)* line->arrow_width * m / 2.0;
+           arrow[2].x = xm - sin(angle)* line->arrow_width * m / 2.0;
+           arrow[2].y = ym + cos(angle)* line->arrow_width * m / 2.0;
            switch(line->arrow_style){
              case GTK_PLOT_SYMBOL_EMPTY:
                gtk_plot_pc_draw_lines (canvas->pc, arrow, 3);
@@ -2559,15 +2772,14 @@ gtk_plot_canvas_draw_child(GtkPlotCanvas *canvas,
           gdk_window_get_size((GdkPixmap *)child->data, &width, &height);
           scale_x = (gdouble)child->allocation.width / (gdouble)width;
           scale_y = (gdouble)child->allocation.height / (gdouble)height;
-          gtk_plot_pc_draw_pixmap(canvas->pc, (GdkPixmap *)child->data, 0, 0,  
+          gtk_plot_pc_draw_pixmap(canvas->pc, (GdkPixmap *)child->data, NULL,
+                                  0, 0,  
                                   child->allocation.x, 
                                   child->allocation.y, 
                                   width, 
                                   height,
                                   scale_x, scale_y);
         } else {
-          GdkColormap *colormap;
-          GdkColor white, black;
           colormap = gdk_colormap_get_system();
           gdk_color_black(colormap, &black);
           gdk_color_white(colormap, &white);
