@@ -38,7 +38,7 @@ static void gtk_plot_csurface_draw_legend	(GtkPlotData *data,
 						 gint x, gint y);
 static void gtk_plot_csurface_draw_private 	(GtkPlotData *data);
 static void gtk_plot_csurface_draw_triangle 	(GtkPlotSurface *surface,
-                                		 GtkPlotVector *points,
+                                		 GtkPlotDTtriangle *triangle,
                                                  gint sign);
 static void gtk_plot_csurface_draw_gradient 	(GtkPlotData *surface,
  						 gint x, gint y);
@@ -46,6 +46,7 @@ static void gtk_plot_csurface_lighting 		(GdkColor *a,
 						 GdkColor *b, 
 						 gdouble normal,
 						 gdouble ambient);
+static void clear_polygons			(GtkPlotCSurface *surface);
 static gint roundint				(gdouble x);
 static void hsv_to_rgb 				(gdouble  h, 
 						 gdouble  s, 
@@ -62,6 +63,16 @@ static void rgb_to_hsv 				(gdouble  r,
 
 static GtkPlotSurfaceClass *parent_class = NULL;
 
+typedef struct
+{
+  GtkPlotPoint p[4];
+  GtkPlotDTtriangle *t;
+  gint n;
+  GdkColor color;
+  gboolean cut_level;
+} GtkPlotPolygon;
+  
+
 GtkType
 gtk_plot_csurface_get_type (void)
 {
@@ -129,9 +140,18 @@ gtk_plot_csurface_init (GtkPlotCSurface *dataset)
   dataset->lines_visible = TRUE;
   dataset->project_xy = TRUE;
   dataset->lines_only = FALSE;
-  dataset->lines_start = NULL; 
-  dataset->lines_end = NULL; 
-  dataset->num_lines = 0;
+  dataset->levels_start = NULL; 
+  dataset->levels_end = NULL; 
+  dataset->sublevels_start = NULL; 
+  dataset->sublevels_end = NULL; 
+  dataset->num_levels = 0;
+  dataset->num_sublevels = 0;
+
+  dataset->polygons = NULL;
+  dataset->npolygons = 0;
+
+  dataset->levels_line = GTK_PLOT_SURFACE(dataset)->mesh_line;
+  dataset->sublevels_line = GTK_PLOT_SURFACE(dataset)->mesh_line;
 }
 
 static void 
@@ -141,15 +161,16 @@ gtk_plot_csurface_destroy 		(GtkObject *object)
 
   surface = GTK_PLOT_CSURFACE(object);
 
-  if(surface->lines_start) g_free(surface->lines_start);
-  if(surface->lines_end) g_free(surface->lines_end);
+  if(surface->levels_start) g_free(surface->levels_start);
+  if(surface->levels_end) g_free(surface->levels_end);
+  if(surface->sublevels_start) g_free(surface->sublevels_start);
+  if(surface->sublevels_end) g_free(surface->sublevels_end);
+
+  clear_polygons(surface);
 
   GTK_OBJECT_CLASS(parent_class)->destroy(object);
 }
 
-
-
-
 GtkWidget*
 gtk_plot_csurface_new (void)
 {
@@ -183,31 +204,93 @@ gtk_plot_csurface_construct_function (GtkPlotCSurface *data,
 static void 
 gtk_plot_csurface_draw_private 	(GtkPlotData *data)
 {
+  GtkPlot *plot;
   GtkPlotSurface *surface;
   GtkPlotCSurface *csurface;
 
   surface = GTK_PLOT_SURFACE(data);
   csurface = GTK_PLOT_CSURFACE(data);
+  plot = data->plot;
 
-  if(csurface->lines_start) g_free(csurface->lines_start); 
-  if(csurface->lines_end) g_free(csurface->lines_end); 
+  if(csurface->levels_start) g_free(csurface->levels_start); 
+  if(csurface->levels_end) g_free(csurface->levels_end); 
+  if(csurface->sublevels_start) g_free(csurface->sublevels_start); 
+  if(csurface->sublevels_end) g_free(csurface->sublevels_end); 
+  clear_polygons(csurface);
+  csurface->polygons = (gpointer)g_new0(GtkPlotPolygon, 1);
 
-  csurface->lines_start = g_new0(GtkPlotPoint, 1);
-  csurface->lines_end = g_new0(GtkPlotPoint, 1);
-  csurface->num_lines = 0;
+  csurface->levels_start = g_new0(GtkPlotVector, 1);
+  csurface->levels_end = g_new0(GtkPlotVector, 1);
+  csurface->sublevels_start = g_new0(GtkPlotVector, 1);
+  csurface->sublevels_end = g_new0(GtkPlotVector, 1);
+  csurface->num_levels = 0;
+  csurface->num_sublevels = 0;
 
   GTK_PLOT_DATA_CLASS(parent_class)->draw_data(data); 
 
+  gtk_plot_set_line_attributes(plot, csurface->levels_line);
+
+  if(!csurface->lines_only){
+    gint i;
+    for(i = 0; i < csurface->npolygons; i++)
+    {
+      GtkPlotPolygon *poly;
+
+      poly = (GtkPlotPolygon *)csurface->polygons[i];
+
+      gtk_plot_pc_set_color(data->plot->pc, &poly->color);
+      gtk_plot_pc_draw_polygon(data->plot->pc, !surface->transparent, poly->p, poly->n);
+    
+      if(csurface->lines_visible || surface->show_mesh || surface->show_grid)
+        gtk_plot_pc_set_color(data->plot->pc, &surface->grid_foreground);
+
+      if(GTK_IS_PLOT3D(data->plot) && csurface->lines_visible && poly->cut_level)
+        gtk_plot_pc_draw_line(data->plot->pc, 
+    	   	              poly->p[0].x, poly->p[0].y, 
+                              poly->p[1].x, poly->p[1].y);
+
+      if(surface->show_mesh || !surface->dt->quadrilateral){
+/*
+        gtk_plot_pc_draw_polygon(data->plot->pc, FALSE, poly->p, poly->n);
+*/
+        gtk_plot_pc_draw_line(data->plot->pc, 
+    	   	              poly->p[1].x, poly->p[1].y, 
+                              poly->p[2].x, poly->p[2].y);
+        if(poly->n == 3)
+          gtk_plot_pc_draw_line(data->plot->pc, 
+    	     	                poly->p[0].x, poly->p[0].y, 
+                                poly->p[2].x, poly->p[2].y);
+        else
+          gtk_plot_pc_draw_line(data->plot->pc, 
+    	     	                poly->p[1].x, poly->p[1].y, 
+                                poly->p[3].x, poly->p[3].y);
+      }
+
+      if(!surface->show_mesh && surface->show_grid && surface->dt->quadrilateral){
+        gtk_plot_pc_draw_line(data->plot->pc, 
+                              poly->t->nb->px, poly->t->nb->py, 
+                              poly->t->nc->px, poly->t->nc->py); 
+      }
+    }
+  }
+
   if(!GTK_IS_PLOT3D(data->plot)){
     gtk_plot_pc_set_color(data->plot->pc, &surface->grid_foreground);
     if(csurface->lines_visible){
       gint i;
-      for(i = 0; i < csurface->num_lines; i++)
+      for(i = 0; i < csurface->num_levels; i++)
          gtk_plot_pc_draw_line(data->plot->pc, 
-                               csurface->lines_start[i].x, 
-                               csurface->lines_start[i].y,
-                               csurface->lines_end[i].x,
-                               csurface->lines_end[i].y);
+                               csurface->levels_start[i].x, 
+                               csurface->levels_start[i].y,
+                               csurface->levels_end[i].x,
+                               csurface->levels_end[i].y);
+      gtk_plot_set_line_attributes(plot, csurface->sublevels_line);
+      for(i = 0; i < csurface->num_sublevels; i++)
+         gtk_plot_pc_draw_line(data->plot->pc, 
+                               csurface->sublevels_start[i].x, 
+                               csurface->sublevels_start[i].y,
+                               csurface->sublevels_end[i].x,
+                               csurface->sublevels_end[i].y);
     }
   }
 
@@ -215,23 +298,24 @@ gtk_plot_csurface_draw_private 	(GtkPlotData *data)
 
 static void
 gtk_plot_csurface_draw_triangle (GtkPlotSurface *surface,
-                                 GtkPlotVector *points,
+                                 GtkPlotDTtriangle *triangle,
                                  gint sign)
 {
   GtkPlot *plot;
   GtkPlotData *data;
   GtkPlotCSurface *csurface;
-  GtkPlotVector side[3], c[2], light, normal;
+  GtkPlotVector points[3], side[3], c[2], light, normal;
   GdkDrawable *drawable;
   GtkPlotPoint t[3], poly[4], proj[2];
+  GtkPlotPolygon *polygon = NULL;
   GdkColor color;
   GdkColor color_h1, color_h2;
   gdouble x, y, z;
-  gint i, n, level;
+  gint i, n, level, sublevel;
   gdouble factor, norm;
   gint mask, vertex[3];
   gdouble min, max, step;
-  gint nlevels;
+  gint nlevels, nsublevels;
   gdouble h, h_next;
 
   csurface = GTK_PLOT_CSURFACE(surface);
@@ -242,7 +326,8 @@ gtk_plot_csurface_draw_triangle (GtkPlotSurface *surface,
   min = data->gradient.begin;
   max = data->gradient.end;
   nlevels = data->gradient.nmajorticks;
-  step = (data->gradient.end - data->gradient.begin) / nlevels;
+  nsublevels = data->gradient.nminorticks;
+  step = (data->gradient.end - data->gradient.begin) / (nlevels * (nsublevels + 1));
   data->gradient.end += step;
 
   light = surface->light;
@@ -251,18 +336,22 @@ gtk_plot_csurface_draw_triangle (GtkPlotSurface *surface,
   light.y /= norm;
   light.z /= norm;
 
-  for(i = 0; i < 3; i++){
-    if(GTK_IS_PLOT3D(plot))
-      gtk_plot3d_get_pixel(GTK_PLOT3D(plot), 
-                           points[i].x, points[i].y, points[i].z,
-                           &x, &y, &z);
-    else
-      gtk_plot_get_pixel(GTK_PLOT(plot), points[i].x, points[i].y, 
-                         &x, &y);
-    
-    t[i].x = x;
-    t[i].y = y;
-  }
+  points[0].x = triangle->na->x;
+  points[0].y = triangle->na->y;
+  points[0].z = triangle->na->z;
+  points[1].x = triangle->nb->x;
+  points[1].y = triangle->nb->y;
+  points[1].z = triangle->nb->z;
+  points[2].x = triangle->nc->x;
+  points[2].y = triangle->nc->y;
+  points[2].z = triangle->nc->z;
+
+  t[0].x = triangle->na->px;
+  t[0].y = triangle->na->py;
+  t[1].x = triangle->nb->px;
+  t[1].y = triangle->nb->py;
+  t[2].x = triangle->nc->px;
+  t[2].y = triangle->nc->py;
 
   side[0].x = (points[1].x - points[0].x) * sign;
   side[0].y = (points[1].y - points[0].y) * sign;
@@ -270,13 +359,17 @@ gtk_plot_csurface_draw_triangle (GtkPlotSurface *surface,
   side[1].x = (points[2].x - points[0].x) * sign;
   side[1].y = (points[2].y - points[0].y) * sign;
   side[1].z = (points[2].z - points[0].z) * sign;
-         
-  normal.x = side[0].y * side[1].z - side[0].z * side[1].y;
-  normal.y = side[0].z * side[1].x - side[0].x * side[1].z;
-  normal.z = side[0].x * side[1].y - side[0].y * side[1].x;
+  
+  if(GTK_IS_PLOT3D(plot)){       
+    normal.x = side[0].y * side[1].z - side[0].z * side[1].y;
+    normal.y = side[0].z * side[1].x - side[0].x * side[1].z;
+    normal.z = side[0].x * side[1].y - side[0].y * side[1].x;
 
-  norm = sqrt(normal.x*normal.x + normal.y*normal.y + normal.z*normal.z); 
-  factor = (normal.x*light.x + normal.y*light.y + normal.z*light.z) / norm;
+    norm = sqrt(normal.x*normal.x + normal.y*normal.y + normal.z*normal.z); 
+    factor = (normal.x*light.x + normal.y*light.y + normal.z*light.z) / norm;
+  } else {
+    factor = 1.0;
+  }
 
   side[0].x = (points[1].x - points[0].x);
   side[0].y = (points[1].y - points[0].y);
@@ -295,9 +388,17 @@ gtk_plot_csurface_draw_triangle (GtkPlotSurface *surface,
     if(!csurface->lines_only){
       gtk_plot_data_get_gradient_level(data, min - step, &color); 
       gtk_plot_csurface_lighting(&color, &color_h1, factor, 1.); 
-      gtk_plot_pc_set_color(plot->pc, &color_h1);
 
-      gtk_plot_pc_draw_polygon(plot->pc, TRUE, t, 3); 
+      polygon = g_new0(GtkPlotPolygon, 1);
+      polygon->n = 3;
+      polygon->color = color_h1;
+      polygon->p[0] = t[0]; 
+      polygon->p[1] = t[1]; 
+      polygon->p[2] = t[2]; 
+      polygon->t = triangle; 
+      csurface->polygons = g_realloc(csurface->polygons, 
+                                    (csurface->npolygons + 1) * sizeof(GtkPlotPolygon));
+      csurface->polygons[csurface->npolygons++] = polygon;
     }
   }
   if(points[0].z > max || points[1].z > max || points[2].z > max){ 
@@ -306,19 +407,38 @@ gtk_plot_csurface_draw_triangle (GtkPlotSurface *surface,
       gtk_plot_csurface_lighting(&color, &color_h1, factor, 1.); 
       gtk_plot_pc_set_color(plot->pc, &color_h1);
 
-      gtk_plot_pc_draw_polygon(plot->pc, TRUE, t, 3); 
+      polygon = g_new0(GtkPlotPolygon, 1);
+      polygon->n = 3;
+      polygon->color = color_h1;
+      polygon->p[0] = t[0]; 
+      polygon->p[1] = t[1]; 
+      polygon->p[2] = t[2]; 
+      polygon->t = triangle; 
+      csurface->polygons = g_realloc(csurface->polygons, 
+                                    (csurface->npolygons + 1) * sizeof(GtkPlotPolygon));
+      csurface->polygons[csurface->npolygons++] = polygon;
     }
   }
 
-  for(level = nlevels + 1; level > 0; level--){
+  for(level = nlevels + 1; level >= 0; level--){
+   for(sublevel = nsublevels + 1; sublevel > 0; sublevel--){
 
     if(points[0].z <= h && points[1].z <= h && points[2].z <= h &&
       (points[0].z > h_next || points[1].z > h_next || points[2].z > h_next)){
       if(!csurface->lines_only){
         gtk_plot_data_get_gradient_level(data, h, &color); 
         gtk_plot_csurface_lighting(&color, &color_h1, factor, 1.); 
-        gtk_plot_pc_set_color(plot->pc, &color_h1);
-        gtk_plot_pc_draw_polygon(plot->pc, TRUE, t, 3); 
+
+        polygon = g_new0(GtkPlotPolygon, 1);
+        polygon->n = 3;
+        polygon->color = color_h1;
+        polygon->p[0] = t[0]; 
+        polygon->p[1] = t[1]; 
+        polygon->p[2] = t[2]; 
+        polygon->t = triangle;
+        csurface->polygons = g_realloc(csurface->polygons, 
+                                    (csurface->npolygons + 1) * sizeof(GtkPlotPolygon));
+        csurface->polygons[csurface->npolygons++] = polygon;
       }
     }
 
@@ -404,66 +524,78 @@ gtk_plot_csurface_draw_triangle (GtkPlotSurface *surface,
           poly[1].y = y;
         }
   
-        if(GTK_IS_PLOT3D(plot) && csurface->project_xy)
+        if(GTK_IS_PLOT3D(plot) && csurface->project_xy){
+          if(sublevel > 1){
+            gtk_plot_set_line_attributes(plot, csurface->sublevels_line);
+          }else{
+            gtk_plot_set_line_attributes(plot, csurface->levels_line);
+          }
+          gtk_plot_pc_set_color(plot->pc, &color_h2);
+ 
           gtk_plot_pc_draw_line(plot->pc, 
    	   	                proj[0].x, proj[0].y, proj[1].x, proj[1].y);
+        }
 
 
         if(!csurface->lines_only){
           if(points[vertex[0]].z <= h_next){
             poly[2] = t[vertex[0]];
-            gtk_plot_pc_draw_polygon(plot->pc, TRUE, poly, 3); 
-/*
-            if(draw_first){
-                GtkPlotPoint poly1[4];
-                poly1[0] = poly[0];
-                poly1[1] = poly[1];
-                poly1[2] = t[vertex[1]];
-                poly1[3] = t[vertex[2]];
-                gtk_plot_data_get_gradient_level(data, h, &color); 
-                gtk_plot_csurface_lighting(&color, &color_h1, factor, 1.); 
-                gtk_plot_pc_set_color(plot->pc, &color_h1);
-                gtk_plot_pc_draw_polygon(plot->pc, TRUE, poly1, 4); 
-                draw_first = FALSE;
-            }
-*/
+
+            polygon = g_new0(GtkPlotPolygon, 1);
+            polygon->n = 3;
+            polygon->color = color_h2;
+            polygon->p[0] = poly[0]; 
+            polygon->p[1] = poly[1]; 
+            polygon->p[2] = poly[2]; 
+            polygon->cut_level = TRUE;
+            polygon->t = triangle; 
+
+            csurface->polygons = g_realloc(csurface->polygons, 
+                                    (csurface->npolygons + 1) * sizeof(GtkPlotPolygon));
+            csurface->polygons[csurface->npolygons++] = polygon;
           }else{
             poly[2] = t[vertex[1]];
             poly[3] = t[vertex[2]];
-            gtk_plot_pc_draw_polygon(plot->pc, TRUE, poly, 4); 
-/*
-            if(draw_first){
-                GtkPlotPoint poly1[3];
-                poly1[0] = poly[0];
-                poly1[1] = poly[1];
-                poly1[2] = t[vertex[0]];
-                gtk_plot_data_get_gradient_level(data, h, &color); 
-                gtk_plot_csurface_lighting(&color, &color_h1, factor, 1.); 
-                gtk_plot_pc_set_color(plot->pc, &color_h1);
-                gtk_plot_pc_draw_polygon(plot->pc, TRUE, poly1, 3); 
-                draw_first = FALSE;
-            }
-*/
+
+            polygon = g_new0(GtkPlotPolygon, 1);
+            polygon->n = 4;
+            polygon->color = color_h2;
+            polygon->p[0] = poly[0]; 
+            polygon->p[1] = poly[1]; 
+            polygon->p[2] = poly[2]; 
+            polygon->p[3] = poly[3]; 
+            polygon->cut_level = TRUE;
+            polygon->t = triangle; 
+
+            csurface->polygons = g_realloc(csurface->polygons, 
+                                    (csurface->npolygons + 1) * sizeof(GtkPlotPolygon));
+            csurface->polygons[csurface->npolygons++] = polygon;
           }
         }
     
-        if(GTK_IS_PLOT3D(plot)){
-          gtk_plot_pc_set_color(plot->pc, &surface->grid_foreground);
-          if(csurface->lines_visible){
-                gtk_plot_pc_draw_line(plot->pc, 
-    	   	                   poly[0].x, poly[0].y, poly[1].x, poly[1].y);
-          }
-        } else {
-          csurface->num_lines++;
-          csurface->lines_start = g_realloc(csurface->lines_start, 
-                                            csurface->num_lines * sizeof(GtkPlotPoint));
-          csurface->lines_end = g_realloc(csurface->lines_end, 
-                                          csurface->num_lines * sizeof(GtkPlotPoint));
+        if(!GTK_IS_PLOT3D(plot)){
+          if(sublevel > 1){
+            csurface->num_sublevels++;
+            csurface->sublevels_start = g_realloc(csurface->sublevels_start, 
+                                            csurface->num_sublevels * sizeof(GtkPlotVector));
+            csurface->sublevels_end = g_realloc(csurface->sublevels_end, 
+                                            csurface->num_sublevels * sizeof(GtkPlotVector));
+            csurface->sublevels_start[csurface->num_sublevels - 1].x = poly[0].x;
+            csurface->sublevels_start[csurface->num_sublevels - 1].y = poly[0].y;
+            csurface->sublevels_end[csurface->num_sublevels - 1].x = poly[1].x;
+            csurface->sublevels_end[csurface->num_sublevels - 1].y = poly[1].y;
 
-          csurface->lines_start[csurface->num_lines - 1].x = poly[0].x;
-          csurface->lines_start[csurface->num_lines - 1].y = poly[0].y;
-          csurface->lines_end[csurface->num_lines - 1].x = poly[1].x;
-          csurface->lines_end[csurface->num_lines - 1].y = poly[1].y;
+          }else{
+            csurface->num_levels++;
+            csurface->levels_start = g_realloc(csurface->levels_start, 
+                                            csurface->num_levels * sizeof(GtkPlotVector));
+            csurface->levels_end = g_realloc(csurface->levels_end, 
+                                            csurface->num_levels * sizeof(GtkPlotVector));
+            csurface->levels_start[csurface->num_levels - 1].x = poly[0].x;
+            csurface->levels_start[csurface->num_levels - 1].y = poly[0].y;
+            csurface->levels_end[csurface->num_levels - 1].x = poly[1].x;
+            csurface->levels_end[csurface->num_levels - 1].y = poly[1].y;
+          }
         }
   
       }
@@ -471,22 +603,23 @@ gtk_plot_csurface_draw_triangle (GtkPlotSurface *surface,
 
     h -= step;
     h_next -= step;
+   }
   }
 
-  if(surface->show_mesh || surface->show_grid)
-    gtk_plot_pc_set_color(plot->pc, &surface->grid_foreground);
 
-  if(surface->show_mesh)
-       gtk_plot_pc_draw_polygon(plot->pc, FALSE, t, 3); 
-
-  if(!surface->show_mesh && surface->show_grid)
-       gtk_plot_pc_draw_line(plot->pc, t[1].x, t[1].y, t[2].x, t[2].y); 
-                              
   data->gradient.begin = min;
   data->gradient.end = max;
   data->gradient.nmajorticks = nlevels;
 }
 
+static void
+clear_polygons(GtkPlotCSurface *surface)
+{
+  if(surface->polygons) g_free(surface->polygons);
+  surface->polygons = NULL;
+  surface->npolygons = 0;
+}
+
 static void
 gtk_plot_csurface_draw_gradient(GtkPlotData *data, gint x, gint y)
 {
@@ -506,7 +639,6 @@ gtk_plot_csurface_draw_gradient(GtkPlotData *data, gint x, gint y)
 
   g_return_if_fail(data->plot != NULL);
   g_return_if_fail(GTK_IS_PLOT(data->plot));
-  g_return_if_fail(GTK_WIDGET_REALIZED(data->plot));
 
   if(!data->show_gradient) return;
 
@@ -550,7 +682,6 @@ gtk_plot_csurface_draw_gradient(GtkPlotData *data, gint x, gint y)
                                  roundint(plot->legends_line_width * m), 
                                  lascent + ldescent);
 
-
       legend.y = (gdouble)(area.y + y + lascent + (lascent + ldescent) / 2) / (gdouble)area.height;
   
       if(level > 1){
@@ -587,7 +718,6 @@ gtk_plot_csurface_get_legend_size(GtkPlotData *data, gint *width, gint *height)
 
   g_return_if_fail(data->plot != NULL);
   g_return_if_fail(GTK_IS_PLOT(data->plot));
-  g_return_if_fail(GTK_WIDGET_REALIZED(data->plot));
 
   plot = data->plot;
 
@@ -645,7 +775,6 @@ gtk_plot_csurface_draw_legend(GtkPlotData *data, gint x, gint y)
 
   g_return_if_fail(data->plot != NULL);
   g_return_if_fail(GTK_IS_PLOT(data->plot));
-  g_return_if_fail(GTK_WIDGET_REALIZED(data->plot));
 
   plot = data->plot;
   area.x = GTK_WIDGET(plot)->allocation.x;
@@ -697,11 +826,11 @@ gtk_plot_csurface_lighting (GdkColor *a, GdkColor *b,
   gdouble red, green, blue;
   gdouble h, s, v;
 
-  normal = MIN(fabs(normal), 1.0);
-
   *b = *a;
   return;
 
+  normal = MIN(fabs(normal), 1.0);
+
   red = a->red;
   green = a->green;
   blue = a->blue;
@@ -877,7 +1006,3 @@ gtk_plot_csurface_labels_visible (GtkPlotCSurface *csurface)
 {
   return (csurface->labels_visible);
 }
-
-
-
-
