@@ -37,6 +37,8 @@ static gchar DEFAULT_FONT[] = "Helvetica";
 static void gtk_plot_data_class_init 		(GtkPlotDataClass *klass);
 static void gtk_plot_data_init 			(GtkPlotData *data);
 static void gtk_plot_data_destroy		(GtkObject *object);
+static void gtk_plot_data_real_update		(GtkPlotData *data, 
+						 gboolean new_range);
 static void gtk_plot_data_draw 			(GtkWidget *widget, 
 						 GdkRectangle *area);
 static void gtk_plot_data_draw_private 		(GtkPlotData *data);
@@ -44,12 +46,12 @@ static void gtk_plot_data_draw_legend		(GtkPlotData *data,
 						 gint x, gint y);
 static void gtk_plot_data_get_legend_size	(GtkPlotData *data, 
 						 gint *width, gint *height);
-static void gtk_plot_data_draw_gradient		(GtkPlotData *data, 
-						 gint x, gint y);
 static void gtk_plot_data_real_draw_gradient	(GtkPlotData *data, 
 						 gint x, gint y);
 static void gtk_plot_data_real_draw		(GtkPlotData *data, 
 						 gint npoints);
+static void gtk_plot_data_real_real_draw	(GtkPlotData *data, 
+						 gint npoints);
 static void gtk_plot_data_real_draw_symbol	(GtkPlotData *data, 
 						 gdouble x, 
                                                  gdouble y, 
@@ -133,8 +135,15 @@ static void rgb_to_hsv 				(gdouble  r,
             					 gdouble *h, 
 						 gdouble *s, 
 						 gdouble *v);
+static void real_autoscale_gradient		(GtkPlotData *data, 
+						 gdouble xmin, gdouble xmax);
+
+static void draw_marker				(GtkPlotData *data, 
+						 GtkPlotMarker *marker);
 
 enum {
+  ADD_TO_PLOT,
+  UPDATE,
   DRAW_DATA,
   LAST_SIGNAL
 };
@@ -142,6 +151,29 @@ enum {
 static GtkWidgetClass *parent_class = NULL;
 static guint data_signals[LAST_SIGNAL] = { 0 };
 
+GtkType
+gtk_plot_marker_get_type (void)
+{
+  static GtkType marker_type = 0;
+
+  if (!marker_type)
+    {
+      GtkTypeInfo data_info =
+      {
+        "GtkPlotMarker",
+        0,
+        0,
+        (GtkClassInitFunc) NULL,
+        (GtkObjectInitFunc) NULL,
+        /* reserved 1*/ NULL,
+        /* reserved 2 */ NULL,
+        (GtkClassInitFunc) NULL,
+      };
+
+      marker_type = gtk_type_unique (GTK_TYPE_BOXED, &data_info);
+    }
+  return marker_type;
+}
 
 GtkType
 gtk_plot_data_get_type (void)
@@ -183,6 +215,8 @@ gtk_plot_data_class_init (GtkPlotDataClass *klass)
   widget_class->draw = gtk_plot_data_draw;
   object_class->destroy = gtk_plot_data_destroy;
 
+  data_class->add_to_plot = NULL;
+  data_class->update = gtk_plot_data_real_update;
   data_class->draw_symbol = gtk_plot_data_real_draw_symbol;
   data_class->draw_data = gtk_plot_data_draw_private;
   data_class->draw_legend = gtk_plot_data_draw_legend;
@@ -191,6 +225,22 @@ gtk_plot_data_class_init (GtkPlotDataClass *klass)
 */
   data_class->get_legend_size = gtk_plot_data_get_legend_size;
 
+  data_signals[ADD_TO_PLOT] =
+    gtk_signal_new ("add_to_plot",
+                    GTK_RUN_LAST,
+                    object_class->type,
+                    GTK_SIGNAL_OFFSET (GtkPlotDataClass, add_to_plot),
+                    gtk_marshal_BOOL__POINTER,
+                    GTK_TYPE_BOOL, 1, GTK_TYPE_PLOT);
+
+  data_signals[UPDATE] =
+    gtk_signal_new ("update",
+                    GTK_RUN_LAST,
+                    object_class->type,
+                    GTK_SIGNAL_OFFSET (GtkPlotDataClass, update),
+                    gtk_marshal_NONE__BOOL,
+                    GTK_TYPE_NONE, 1, GTK_TYPE_BOOL);
+
   data_signals[DRAW_DATA] =
     gtk_signal_new ("draw_data",
                     GTK_RUN_FIRST,
@@ -203,7 +253,6 @@ gtk_plot_data_class_init (GtkPlotDataClass *klass)
 
 }
 
-
 static void
 gtk_plot_data_init (GtkPlotData *dataset)
 {
@@ -230,8 +279,11 @@ gtk_plot_data_init (GtkPlotData *dataset)
   dataset->gradient.begin = 0.0;
   dataset->gradient.end = 1.0;
   dataset->gradient.nmajorticks = 10;
+  dataset->gradient.nminorticks = 0;
   dataset->gradient.major = NULL;
   dataset->gradient.major_values = NULL;
+  dataset->gradient.minor = NULL;
+  dataset->gradient.minor_values = NULL;
   dataset->gradient_mask = GTK_PLOT_GRADIENT_H;
   dataset->show_gradient = FALSE;
   gtk_plot_data_calc_gradient(dataset);
@@ -240,10 +292,13 @@ gtk_plot_data_init (GtkPlotData *dataset)
   dataset->plot = NULL;
 
   dataset->is_function = FALSE;
+  dataset->is_iterator = FALSE;
+  dataset->iterator_mask = GTK_PLOT_DATA_X|GTK_PLOT_DATA_Y;
   dataset->show_legend = TRUE;
   dataset->show_labels = FALSE;
   dataset->fill_area = FALSE;
   dataset->function = NULL;
+  dataset->iterator = NULL;
   dataset->num_points = 0;
   dataset->x = NULL;
   dataset->y = NULL;
@@ -253,6 +308,8 @@ gtk_plot_data_init (GtkPlotData *dataset)
   dataset->dy = NULL;
   dataset->dz = NULL;
   dataset->da = NULL;
+  dataset->a_scale = 1.0;
+
   dataset->labels = NULL;
   dataset->x_step = .5;
   dataset->y_step = .5;
@@ -302,6 +359,9 @@ gtk_plot_data_init (GtkPlotData *dataset)
 
   dataset->link = NULL;
 
+  dataset->markers = NULL;
+  dataset->show_markers = TRUE;
+
   gtk_plot_data_labels_set_attributes(dataset, 
                                       DEFAULT_FONT,
                                       DEFAULT_FONT_HEIGHT,
@@ -310,6 +370,8 @@ gtk_plot_data_init (GtkPlotData *dataset)
                                       &white);
 
   gtk_psfont_init();
+
+  dataset->redraw_pending = TRUE;
 }
 
 static void
@@ -328,8 +390,18 @@ gtk_plot_data_destroy (GtkObject *object)
 
   if(data->gradient.major){
      g_free(data->gradient.major);
+  }
+  if(data->gradient.major_values){
      g_free(data->gradient.major_values);
   }
+  if(data->gradient.minor){
+     g_free(data->gradient.minor);
+  }
+  if(data->gradient.minor_values){
+     g_free(data->gradient.minor_values);
+  }
+
+  gtk_plot_data_remove_markers(data);
 
   gtk_psfont_unref();
 }
@@ -364,12 +436,97 @@ gtk_plot_data_construct_function (GtkPlotData *data, GtkPlotFunc function)
   data->function = function;
 }
 
+GtkWidget*
+gtk_plot_data_new_iterator (GtkPlotIterator iterator, gint npoints, guint16 mask)
+{
+  GtkWidget *dataset;
+
+  dataset = gtk_type_new (gtk_plot_data_get_type ());
+
+  gtk_plot_data_construct_iterator (GTK_PLOT_DATA(dataset), iterator, npoints, mask);
+
+  return (dataset);
+}
+
+void
+gtk_plot_data_construct_iterator (GtkPlotData *data, 
+				  GtkPlotIterator iterator, gint npoints, guint16 mask)
+{
+  data->is_iterator = TRUE;
+  data->iterator_mask = mask;
+  data->iterator = iterator;
+  data->num_points = npoints; 
+}
+
+void
+gtk_plot_data_update(GtkPlotData *data)
+{
+  GTK_PLOT_DATA_CLASS(GTK_OBJECT(data)->klass)->update(data, TRUE);
+}
+
+static void
+gtk_plot_data_real_update(GtkPlotData *data, gboolean new_range)
+{
+  data->redraw_pending = TRUE;
+}
+
 static void
 gtk_plot_data_draw (GtkWidget *widget, GdkRectangle *area)
 {
   if(!GTK_WIDGET_VISIBLE(widget)) return;
   gtk_signal_emit(GTK_OBJECT(widget), data_signals[DRAW_DATA], NULL);
   GTK_PLOT_DATA_CLASS(GTK_OBJECT(widget)->klass)->draw_data(GTK_PLOT_DATA(widget));
+
+  GTK_PLOT_DATA(widget)->redraw_pending = FALSE;
+}
+
+static void
+draw_marker(GtkPlotData *data, GtkPlotMarker *marker)
+{
+  GtkPlot *plot;
+  GdkColor black;
+  GdkColormap *colormap = NULL;
+  GtkPlotPoint p[6];
+  gdouble x, y, z, a, dx, dy, dz, da;
+  gchar *label;
+  gboolean error;
+  gdouble px, py;
+
+  if(!data->plot) return;
+  plot = data->plot;
+
+  gtk_plot_data_get_point(data, marker->point, 
+			  &x, &y, &z, &a, &dx, &dy, &dz, &da,  
+			  &label, &error);
+
+  if(x < plot->xmin || y < plot->ymin || x >= plot->xmax || y >= plot->ymax) 
+     return;
+
+  gtk_plot_get_pixel(plot, x, y, &px, &py);
+
+  colormap = gdk_colormap_get_system();
+  gdk_color_black(colormap, &black);
+  gtk_plot_pc_set_color(plot->pc, &black);
+  gtk_plot_pc_set_lineattr(plot->pc, 1, 0, 0, 0);
+
+  p[0].x = px;
+  p[0].y = py;
+  p[1].x = px - 6;
+  p[1].y = py - 12;
+  p[2].x = px + 6;
+  p[2].y = py - 12;
+  gtk_plot_pc_draw_polygon(plot->pc, TRUE, p, 3);
+  p[1].x = px - 6;
+  p[1].y = py + 12;
+  p[2].x = px + 6;
+  p[2].y = py + 12;
+  gtk_plot_pc_draw_polygon(plot->pc, TRUE, p, 3);
+  p[0].x = px - 6;
+  p[0].y = py;
+  p[1].x = px + 7;
+  p[1].y = py;
+  gtk_plot_pc_set_lineattr(plot->pc, 3, 0, 0, 0);
+  gtk_plot_pc_draw_polygon(plot->pc, FALSE, p, 2);
 }
 
 void
@@ -378,12 +535,27 @@ gtk_plot_data_paint (GtkPlotData *data)
   if(!GTK_WIDGET_VISIBLE(GTK_WIDGET(data))) return;
 
   gtk_signal_emit(GTK_OBJECT(data), data_signals[DRAW_DATA], NULL);
+
+  data->redraw_pending = FALSE;
 }
 
 static void
 gtk_plot_data_draw_private (GtkPlotData *data)
 {
+  GList *list = NULL;
+
   gtk_plot_data_real_draw(data, data->num_points);
+
+  if(data->show_markers){
+    list = data->markers;
+    while(list){
+      GtkPlotMarker *marker;
+
+      marker = (GtkPlotMarker *)list->data;
+      draw_marker(data, marker); 
+      list = list->next;
+    } 
+  }
 }
 
 static void
@@ -397,9 +569,15 @@ gtk_plot_data_real_draw   (GtkPlotData *dataset,
   GdkRectangle area, clip_area;
   gdouble x, y, z = 0., a = 0.;
   gdouble dx = 0., dy = 0., dz = 0., da = 0.;
-  gint n;
   gdouble *fx = NULL;
   gdouble *fy = NULL;
+  gdouble *fz = NULL;
+  gdouble *fa = NULL;
+  gdouble *fdx = NULL;
+  gdouble *fdy = NULL;
+  gdouble *fdz = NULL;
+  gdouble *fda = NULL;
+  gchar **fl = NULL;
   gboolean error;
   gdouble m;
 
@@ -419,7 +597,6 @@ gtk_plot_data_real_draw   (GtkPlotData *dataset,
   gtk_plot_pc_gsave(plot->pc);
 
   m = plot->magnification;
-  npoints = MIN(npoints, dataset->num_points);
 
   area.x = widget->allocation.x;
   area.y = widget->allocation.y;
@@ -431,69 +608,10 @@ gtk_plot_data_real_draw   (GtkPlotData *dataset,
   clip_area.width = roundint(plot->width * area.width);
   clip_area.height = roundint(plot->height * area.height);
 
-  if(!dataset->is_function)
-    {
-       gtk_plot_data_connect_points (dataset, npoints+1);
-       gtk_plot_data_draw_xyz(dataset, npoints);
-       for(n=dataset->num_points-npoints; n<=dataset->num_points-1; n++)
-         {
-           x = dataset->x[n];
-           y = dataset->y[n];
-           if(dataset->z != NULL) z = dataset->z[n];
-           if(dataset->a != NULL) a = dataset->a[n];
-           if(dataset->dx != NULL) dx = dataset->dx[n];
-           if(dataset->dy != NULL) dy = dataset->dy[n];
-           if(dataset->dz != NULL) dz = dataset->dz[n];
-           if(dataset->da != NULL) da = dataset->da[n];
-
-           if(!plot->clip_data || (plot->clip_data && x >= plot->xmin && x <= plot->xmax)){
-               GdkColor symbol_color, border_color;
-               gint symbol_size;
-
-               symbol_color = dataset->symbol.color;
-               border_color = dataset->symbol.border.color;
-               symbol_size = dataset->symbol.size;
-               if(dataset->da){
-                  GdkColor level_color;
-                  gtk_plot_data_get_gradient_level(dataset, da, &level_color);
-                  dataset->symbol.color = level_color;
-                  dataset->symbol.border.color = level_color;
-               }
-               if(dataset->a && !GTK_IS_PLOT3D(plot)){
-                  gdouble px, py, px0, py0;     
-                  gtk_plot_get_pixel(plot, x, y, &px, &py);
-                  gtk_plot_get_pixel(plot, x + a, y, &px0, &py0);
-                  dataset->symbol.size = fabs((px - px0)/plot->magnification);
-               }
-               GTK_PLOT_DATA_CLASS(GTK_OBJECT(dataset)->klass)->draw_symbol(dataset, x, y, z, a, dx, dy, dz, da); 
-               dataset->symbol.color = symbol_color;
-               dataset->symbol.border.color = border_color;
-               dataset->symbol.size = symbol_size;
-
-               if(dataset->show_labels){
-                 if(dataset->labels && dataset->labels[n]){
-                    GtkPlotText label;
-                    gdouble px, py, pz;
-
-                    if(GTK_IS_PLOT3D(plot))
-                        gtk_plot3d_get_pixel(GTK_PLOT3D(plot), x, y, z, 
-                                             &px, &py, &pz);
-                    else
-                        gtk_plot_get_pixel(plot, x, y, &px, &py);
-
-                    label = dataset->labels_attr;
-                    label.text = dataset->labels[n];
-                    label.x = (gdouble)px / (gdouble)area.width;
-                    label.y = (gdouble)(py - roundint((dataset->labels_offset + dataset->symbol.size) * m)) / (gdouble)area.height; 
-                    gtk_plot_draw_text(plot, label);	
-                 }          
-               }
-           }
-         } 
-    } 
-  else
+  if(dataset->is_function)
     {
        gdouble xmin, xmax, px, py;
+
        function = dataset;
        function->num_points = 0;
        gtk_plot_get_pixel(plot, plot->xmin, 0, &xmin, &py);
@@ -532,9 +650,193 @@ gtk_plot_data_real_draw   (GtkPlotData *dataset,
        function->x = NULL;
        function->y = NULL;
     }
+  else if(dataset->is_iterator)
+    {
+       gint iter;
+       gchar *label;
+   
+       function = dataset;
+       if(function->iterator_mask & GTK_PLOT_DATA_X)
+         fx = g_new0(gdouble, npoints);
+       if(function->iterator_mask & GTK_PLOT_DATA_Y)
+         fy = g_new0(gdouble, npoints); 
+       if(function->iterator_mask & GTK_PLOT_DATA_Z)
+         fz = g_new0(gdouble, npoints);
+       if(function->iterator_mask & GTK_PLOT_DATA_A)
+         fa = g_new0(gdouble, npoints); 
+       if(function->iterator_mask & GTK_PLOT_DATA_DX)
+         fdx = g_new0(gdouble, npoints);
+       if(function->iterator_mask & GTK_PLOT_DATA_DY)
+         fdy = g_new0(gdouble, npoints); 
+       if(function->iterator_mask & GTK_PLOT_DATA_DZ)
+         fdz = g_new0(gdouble, npoints);
+       if(function->iterator_mask & GTK_PLOT_DATA_DA)
+         fda = g_new0(gdouble, npoints);
+       if(function->iterator_mask & GTK_PLOT_DATA_LABELS)
+         fl = (gchar **)g_malloc0(npoints * sizeof(gchar *)); 
+
+       for(iter = 0; iter < npoints; iter++){
+            function->iterator (plot, dataset, 
+                                function->num_points - npoints + iter, 
+                                &x, &y, &z, &a, &dx, &dy, &dz, &da, &label, &error);
+
+            if(error)
+              {
+                 break;
+              }
+            else
+              {
+                if(function->iterator_mask & GTK_PLOT_DATA_X) fx[iter] = x;
+                if(function->iterator_mask & GTK_PLOT_DATA_Y) fy[iter] = y;
+                if(function->iterator_mask & GTK_PLOT_DATA_Z) fz[iter] = z;
+                if(function->iterator_mask & GTK_PLOT_DATA_A) fa[iter] = a;
+                if(function->iterator_mask & GTK_PLOT_DATA_DX) fdx[iter] = dx;
+                if(function->iterator_mask & GTK_PLOT_DATA_DY) fdy[iter] = dy;
+                if(function->iterator_mask & GTK_PLOT_DATA_DZ) fdz[iter] = dz;
+                if(function->iterator_mask & GTK_PLOT_DATA_DA) fda[iter] = da;
+                if(function->iterator_mask & GTK_PLOT_DATA_LABELS) 
+                                                   fl[iter] = g_strdup(label);
+              }
+       }
+       function->x = fx;
+       function->y = fy;
+       function->z = fz;
+       function->a = fa;
+       function->dx = fdx;
+       function->dy = fdy;
+       function->dz = fdz;
+       function->da = fda;
+       function->labels = fl;
+       gtk_plot_data_real_real_draw(function, MIN(iter, npoints));
+
+       if(fx) g_free(fx);
+       if(fy) g_free(fy);
+       if(fz) g_free(fz);
+       if(fa) g_free(fa);
+       if(fdx) g_free(fdx);
+       if(fdy) g_free(fdy);
+       if(fdz) g_free(fdz);
+       if(fda) g_free(fda);
+       if(fl) {
+         for(iter = 0; iter < npoints; iter++) g_free(fl[iter]);
+         g_free(fl);
+       }
+       function->x = NULL;
+       function->y = NULL;
+       function->z = NULL;
+       function->a = NULL;
+       function->dx = NULL;
+       function->dy = NULL;
+       function->dz = NULL;
+       function->da = NULL;
+       function->labels = NULL;
+    }
+  else
+       gtk_plot_data_real_real_draw(dataset, npoints);
+
   gtk_plot_pc_grestore(plot->pc);
 }
 
+static void
+gtk_plot_data_real_real_draw   (GtkPlotData *dataset,  
+			   	gint npoints)
+{
+  GtkWidget *widget;
+  GtkPlot *plot = NULL;
+  GdkRectangle area;
+  GdkColormap *colormap;
+  gdouble x, y, z = 0., a = 0.;
+  gdouble dx = 0., dy = 0., dz = 0., da = 0.;
+  gint n;
+  gdouble m;
+
+  g_return_if_fail(GTK_IS_PLOT_DATA(dataset));
+  g_return_if_fail(dataset->plot != NULL);
+  g_return_if_fail(GTK_IS_PLOT(dataset->plot));
+  g_return_if_fail(GTK_WIDGET_REALIZED(dataset->plot));
+
+  plot = dataset->plot;
+  widget = GTK_WIDGET(plot);
+  colormap = gdk_colormap_get_system();
+
+  if(!GTK_WIDGET_DRAWABLE(widget)) return;
+  if(!GTK_WIDGET_VISIBLE(widget)) return;
+  if(!GTK_WIDGET_VISIBLE(GTK_WIDGET(dataset))) return;
+
+  area.x = widget->allocation.x;
+  area.y = widget->allocation.y;
+  area.width = widget->allocation.width;
+  area.height = widget->allocation.height;
+
+  gtk_plot_pc_gsave(plot->pc);
+
+  m = plot->magnification;
+  npoints = MIN(npoints, dataset->num_points);
+
+  gtk_plot_data_connect_points (dataset, npoints+1);
+  gtk_plot_data_draw_xyz(dataset, npoints);
+  for(n=dataset->num_points-npoints; n<=dataset->num_points-1; n++)
+    {
+      x = dataset->x[n];
+      y = dataset->y[n];
+      if(dataset->z != NULL) z = dataset->z[n];
+      if(dataset->a != NULL) a = dataset->a[n];
+      if(dataset->dx != NULL) dx = dataset->dx[n];
+      if(dataset->dy != NULL) dy = dataset->dy[n];
+      if(dataset->dz != NULL) dz = dataset->dz[n];
+      if(dataset->da != NULL) da = dataset->da[n];
+
+      if(!plot->clip_data || 
+        (plot->clip_data && GTK_IS_PLOT_POLAR(plot) && x >= GTK_PLOT_POLAR(plot)->r->min && x <= GTK_PLOT_POLAR(plot)->r->max) || 
+        (plot->clip_data && x >= plot->xmin && x <= plot->xmax)){
+          GdkColor symbol_color, border_color;
+          gint symbol_size;
+
+          symbol_color = dataset->symbol.color;
+          border_color = dataset->symbol.border.color;
+          symbol_size = dataset->symbol.size;
+          if(dataset->da){
+             GdkColor level_color;
+             gtk_plot_data_get_gradient_level(dataset, da, &level_color);
+             dataset->symbol.color = level_color;
+             dataset->symbol.border.color = level_color;
+          }
+          if(dataset->a && !GTK_IS_PLOT3D(plot)){
+             gdouble px, py, px0, py0;     
+             gtk_plot_get_pixel(plot, x, y, &px, &py);
+             gtk_plot_get_pixel(plot, x + a, y, &px0, &py0);
+             dataset->symbol.size = fabs((px - px0)/plot->magnification);
+             dataset->symbol.size *= fabs(dataset->a_scale);
+          }
+          GTK_PLOT_DATA_CLASS(GTK_OBJECT(dataset)->klass)->draw_symbol(dataset, x, y, z, a, dx, dy, dz, da); 
+          dataset->symbol.color = symbol_color;
+          dataset->symbol.border.color = border_color;
+          dataset->symbol.size = symbol_size;
+
+          if(dataset->show_labels){
+            if(dataset->labels && dataset->labels[n]){
+               GtkPlotText label;
+               gdouble px, py, pz;
+
+               if(GTK_IS_PLOT3D(plot))
+                   gtk_plot3d_get_pixel(GTK_PLOT3D(plot), x, y, z, 
+                                        &px, &py, &pz);
+               else
+                   gtk_plot_get_pixel(plot, x, y, &px, &py);
+
+               label = dataset->labels_attr;
+               label.text = dataset->labels[n];
+               label.x = (gdouble)px / (gdouble)area.width;
+               label.y = (gdouble)(py - roundint((dataset->labels_offset + dataset->symbol.size) * m)) / (gdouble)area.height; 
+               gtk_plot_draw_text(plot, label);	
+            }          
+          }
+      }
+    } 
+ 
+}
+
+
 static void
 gtk_plot_data_draw_legend(GtkPlotData *data, gint x, gint y)
 {
@@ -618,7 +920,7 @@ gtk_plot_data_draw_legend(GtkPlotData *data, gint x, gint y)
   gtk_plot_pc_grestore(plot->pc);
 }
 
-static void
+void
 gtk_plot_data_draw_gradient(GtkPlotData *data, gint x, gint y)
 {
   gtk_plot_data_real_draw_gradient(data, x, y);
@@ -784,13 +1086,18 @@ gtk_plot_data_real_draw_symbol (GtkPlotData *data,
   plot = data->plot;
 
   if(GTK_IS_PLOT_POLAR(plot)){
+    GtkPlotPolar *polar = GTK_PLOT_POLAR(plot);
     if(plot->clip_data &&
-       (y < plot->xmin || y > plot->xmax || x < plot->ymin || x > plot->ymax)) 
+       (x < polar->r->min || x > polar->r->max || y < polar->angle->min || y > polar->angle->max)) 
              return; 
-  } else {
+  } else if(GTK_IS_PLOT3D(plot)){
     if(plot->clip_data && (x < plot->xmin || x > plot->xmax)) return;
     if(plot->clip_data && data->symbol.symbol_type != GTK_PLOT_SYMBOL_IMPULSE &&
        (y < plot->ymin || y > plot->ymax)) return; 
+  } else {
+    if(plot->clip_data &&
+       (x < plot->xmin || x > plot->xmax || y < plot->ymin || y > plot->ymax)) 
+             return; 
   }
 
   if(GTK_IS_PLOT3D(plot))
@@ -1053,17 +1360,19 @@ gtk_plot_data_draw_errbars(GtkPlotData *dataset,
               errbar[0].y = eu_y; 
               errbar[1].x = px + m * dataset->zerrbar_caps/2.; 
               errbar[1].y = eu_y; 
-              errbar[2].x = px; 
-     
-              errbar[2].y = eu_y; 
-              errbar[3].x = px; 
-     
-              errbar[3].y = ed_y; 
-              errbar[4].x = px - m * dataset->zerrbar_caps/2.; 
-              errbar[4].y = ed_y; 
-              errbar[5].x = px + m * dataset->zerrbar_caps/2.; 
-              errbar[5].y = ed_y; 
-              gtk_plot_pc_draw_lines(plot->pc, errbar, 6);
+              gtk_plot_pc_draw_lines(plot->pc, errbar, 2);
+
+              errbar[0].x = px; 
+              errbar[0].y = eu_y; 
+              errbar[1].x = px; 
+              errbar[1].y = ed_y; 
+              gtk_plot_pc_draw_lines(plot->pc, errbar, 2);
+
+              errbar[0].x = px - m * dataset->zerrbar_caps/2.; 
+              errbar[0].y = ed_y; 
+              errbar[1].x = px + m * dataset->zerrbar_caps/2.; 
+              errbar[1].y = ed_y; 
+              gtk_plot_pc_draw_lines(plot->pc, errbar, 2);
            }
   } else {
        gtk_plot_get_pixel(plot, x, y, &px, &py);
@@ -1078,17 +1387,19 @@ gtk_plot_data_draw_errbars(GtkPlotData *dataset,
               errbar[0].y = py - m * dataset->xerrbar_caps/2.; 
               errbar[1].x = el_x; 
               errbar[1].y = py + m * dataset->xerrbar_caps/2.; 
-              errbar[2].x = el_x; 
-              errbar[2].y = py; 
+              gtk_plot_pc_draw_lines(plot->pc, errbar, 2);
 
-              errbar[3].x = er_x; 
-              errbar[3].y = py; 
+              errbar[0].x = el_x; 
+              errbar[0].y = py; 
+              errbar[1].x = er_x; 
+              errbar[1].y = py; 
+              gtk_plot_pc_draw_lines(plot->pc, errbar, 2);
      
-              errbar[4].x = er_x; 
-              errbar[4].y = py - m * dataset->xerrbar_caps/2.; 
-              errbar[5].x = er_x; 
-              errbar[5].y = py + m * dataset->xerrbar_caps/2.; 
-              gtk_plot_pc_draw_lines(plot->pc, errbar, 6);
+              errbar[0].x = er_x; 
+              errbar[0].y = py - m * dataset->xerrbar_caps/2.; 
+              errbar[1].x = er_x; 
+              errbar[1].y = py + m * dataset->xerrbar_caps/2.; 
+              gtk_plot_pc_draw_lines(plot->pc, errbar, 2);
            }
 
        if(dataset->show_yerrbars)
@@ -1097,17 +1408,19 @@ gtk_plot_data_draw_errbars(GtkPlotData *dataset,
               errbar[0].y = eu_y; 
               errbar[1].x = px + m * dataset->yerrbar_caps/2.; 
               errbar[1].y = eu_y; 
-              errbar[2].x = px; 
-     
-              errbar[2].y = eu_y; 
-              errbar[3].x = px; 
-     
-              errbar[3].y = ed_y; 
-              errbar[4].x = px - m * dataset->yerrbar_caps/2.; 
-              errbar[4].y = ed_y; 
-              errbar[5].x = px + m * dataset->yerrbar_caps/2.; 
-              errbar[5].y = ed_y; 
-              gtk_plot_pc_draw_lines(plot->pc, errbar, 6);
+              gtk_plot_pc_draw_lines(plot->pc, errbar, 2);
+
+              errbar[0].x = px; 
+              errbar[0].y = eu_y; 
+              errbar[1].x = px; 
+              errbar[1].y = ed_y; 
+              gtk_plot_pc_draw_lines(plot->pc, errbar, 2);
+
+              errbar[0].x = px - m * dataset->yerrbar_caps/2.; 
+              errbar[0].y = ed_y; 
+              errbar[1].x = px + m * dataset->yerrbar_caps/2.; 
+              errbar[1].y = ed_y; 
+              gtk_plot_pc_draw_lines(plot->pc, errbar, 2);
            }
   }
 }
@@ -1588,6 +1901,44 @@ gtk_plot_data_get_points(GtkPlotData *data,
   *num_points = data->num_points;
 }
 
+void
+gtk_plot_data_get_point(GtkPlotData *dataset, gint n, 
+                        gdouble *x, gdouble *y, gdouble *z, gdouble *a, 
+                        gdouble *dx, gdouble *dy, gdouble *dz, gdouble *da, 
+                        gchar **label, gboolean *error)
+{
+  *error = FALSE;
+
+  if(dataset->is_function){
+    g_warning("This functions does not work for functions");
+    *error = TRUE;
+    return;
+  } else if(dataset->is_iterator) {
+    if(n >= dataset->num_points){
+       g_warning("n >= dataset->num_points");
+       *error = TRUE;
+       return;
+    }
+    dataset->iterator (GTK_PLOT(dataset->plot), dataset, n, 
+                       x, y, z, a, dx, dy, dz, da, label, error);
+  } else {
+    if(n >= dataset->num_points){
+       g_warning("n >= dataset->num_points");
+       *error = TRUE;
+       return;
+    }
+    if(dataset->x) *x = dataset->x[n]; 
+    if(dataset->y) *y = dataset->y[n]; 
+    if(dataset->z) *z = dataset->z[n]; 
+    if(dataset->a) *a = dataset->a[n]; 
+    if(dataset->dx) *dx = dataset->dx[n]; 
+    if(dataset->dy) *dy = dataset->dy[n]; 
+    if(dataset->dz) *dz = dataset->dz[n]; 
+    if(dataset->da) *da = dataset->da[n]; 
+    if(dataset->labels) *label = dataset->labels[n]; 
+  }
+}
+
 void
 gtk_plot_data_set_x(GtkPlotData *data, 
                     gdouble *x) 
@@ -1647,6 +1998,12 @@ gtk_plot_data_set_da(GtkPlotData *data,
   data->da = da;
 }
 
+void
+gtk_plot_data_set_a_scale(GtkPlotData *data, gdouble a_scale) 
+{
+  data->a_scale = a_scale;
+}
+
 
 void
 gtk_plot_data_set_labels(GtkPlotData *data, 
@@ -1655,6 +2012,7 @@ gtk_plot_data_set_labels(GtkPlotData *data,
   data->labels = labels;
 }
 
+
 gdouble *
 gtk_plot_data_get_x(GtkPlotData *dataset, gint *num_points)
 {
@@ -1711,6 +2069,12 @@ gtk_plot_data_get_da(GtkPlotData *dataset, gint *num_points)
   return(dataset->da);
 }
 
+gdouble 
+gtk_plot_data_get_a_scale(GtkPlotData *dataset)
+{
+  return(dataset->a_scale);
+}
+
 
 gchar **
 gtk_plot_data_get_labels(GtkPlotData *dataset, gboolean *show_labels)
@@ -2001,6 +2365,143 @@ gtk_plot_data_gradient_visible (GtkPlotData *data)
   return(data->show_gradient);
 }
 
+void
+gtk_plot_data_gradient_autoscale_a (GtkPlotData *data)
+{
+  gdouble amin, amax;
+  gint n;
+  gboolean change = FALSE;
+
+  if(data->is_function) return;
+  if(!data->a && !(data->iterator_mask & GTK_PLOT_DATA_A)) return;
+
+  amin = 1.e16;
+  amax = -1.e16;
+
+  for(n = 0; n < data->num_points; n++){
+      gdouble fx, fy, fz, fa;
+      gdouble fdx, fdy, fdz, fda;
+      gchar *label;
+      gboolean error;
+      gtk_plot_data_get_point(data, n,
+                              &fx, &fy, &fz, &fa,
+                              &fdx, &fdy, &fdz, &fda,
+                              &label, &error);
+      if(fa < amin) amin = fa;
+      if(fa > amax) amax = fa;
+      change = TRUE;
+  }
+
+  if(!change) return;
+  real_autoscale_gradient(data, amin, amax);
+}
+
+void
+gtk_plot_data_gradient_autoscale_da (GtkPlotData *data)
+{
+  gdouble amin, amax;
+  gint n;
+  gboolean change = FALSE;
+
+  if(data->is_function) return;
+  if(!data->da && !(data->iterator_mask & GTK_PLOT_DATA_DA)) return;
+
+  amin = 1.e16;
+  amax = -1.e16;
+
+  for(n = 0; n < data->num_points; n++){
+      gdouble fx, fy, fz, fa;
+      gdouble fdx, fdy, fdz, fda;
+      gchar *label;
+      gboolean error;
+      gtk_plot_data_get_point(data, n,
+                              &fx, &fy, &fz, &fa,
+                              &fdx, &fdy, &fdz, &fda,
+                              &label, &error);
+      if(fda < amin) amin = fda;
+      if(fda > amax) amax = fda;
+      change = TRUE;
+  }
+
+  if(!change) return;
+  real_autoscale_gradient(data, amin, amax);
+}
+
+void
+gtk_plot_data_gradient_autoscale_z (GtkPlotData *data)
+{
+  gdouble zmin, zmax;
+  gint n;
+  gboolean change = FALSE;
+
+  if(data->is_function) return;
+  if(!data->z && !(data->iterator_mask & GTK_PLOT_DATA_Z)) return;
+
+  zmin = 1.e16;
+  zmax = -1.e16;
+
+  for(n = 0; n < data->num_points; n++){
+      gdouble fx, fy, fz, fa;
+      gdouble fdx, fdy, fdz, fda;
+      gchar *label;
+      gboolean error;
+      gtk_plot_data_get_point(data, n,
+                              &fx, &fy, &fz, &fa,
+                              &fdx, &fdy, &fdz, &fda,
+                              &label, &error);
+      if(fz < zmin) zmin = fz;
+      if(fz > zmax) zmax = fz;
+      change = TRUE;
+  }
+
+  if(!change) return;
+  real_autoscale_gradient(data, zmin, zmax);
+}
+
+static void
+real_autoscale_gradient(GtkPlotData *data, gdouble xmin, gdouble xmax)
+{
+  gdouble dx;
+  gdouble pmin, pmax;
+  gdouble pstep;
+  gdouble min, max;
+
+  min = xmin;
+  max = xmax;
+  if(xmin == xmax){
+    if(xmin == 0.0){
+      max = xmax = 0.1;
+    }else{
+      pstep = floor(log10(fabs(xmin)));
+      dx = xmin/pow(10., pstep) * pow(10., pstep);
+      max = xmax = xmin + dx;
+      min = xmin = xmin - dx;
+    }
+  }
+  dx = (xmax - xmin)/10.;
+  xmin += dx; 
+  xmax -= dx; 
+  if(xmin == 0.0) xmin -= dx; 
+  if(xmax == 0.0) xmax += dx; 
+  pmin = floor(log10(fabs(xmin))) - 1.;
+  pmax = floor(log10(fabs(xmax))) - 1.;
+  xmin = floor(xmin/pow(10., pmin)) * pow(10., pmin);
+  xmax = floor(xmax/pow(10., pmax)) * pow(10., pmax);
+  pstep = floor(log10(fabs(dx)));
+  dx = data->gradient.step = floor(dx/pow(10., pstep)) * pow(10., pstep);
+  while(xmin >= min) xmin -= dx;
+  while(xmax <= max) xmax += dx;
+  dx = floor((xmax - xmin)/data->gradient.step);
+  while(dx > 10.){
+    data->gradient.step *= 2;
+    dx = floor((xmax - xmin)/data->gradient.step);
+  }
+  xmin = floor(xmin/data->gradient.step) * data->gradient.step;
+  xmax = ceil(xmax/data->gradient.step) * data->gradient.step;
+
+  gtk_plot_data_set_gradient(data, xmin, xmax, 8, 0);
+}
+
 void
 gtk_plot_data_set_gradient_colors (GtkPlotData *data,
                                    const GdkColor *min,
@@ -2021,22 +2522,26 @@ gtk_plot_data_get_gradient_colors (GtkPlotData *data,
 
 void
 gtk_plot_data_set_gradient (GtkPlotData *data,
-                            gdouble min, gdouble max, gint nlevels)
+                            gdouble min, gdouble max, 
+			    gint nlevels, gint nsublevels)
 {
   data->gradient.begin = min;
   data->gradient.end = max;
   data->gradient.nmajorticks = nlevels;
+  data->gradient.nminorticks = nsublevels;
 
   gtk_plot_data_calc_gradient(data);
 }
 
 void
 gtk_plot_data_get_gradient (GtkPlotData *data,
-                            gdouble *min, gdouble *max, gint *nlevels)
+                            gdouble *min, gdouble *max, 
+			    gint *nlevels, gint *nsublevels)
 {
   *min = data->gradient.begin;
   *max = data->gradient.end;
   *nlevels = data->gradient.nmajorticks;
+  *nsublevels = data->gradient.nminorticks;
 }
 
 void
@@ -2265,8 +2770,9 @@ gtk_plot_data_calc_gradient(GtkPlotData *data)
   gdouble min = 0., max = 0.;
   gdouble absmin = 0., absmax = 0.;
   gdouble tick;
-  gdouble major_step;
-  gint nmajor;
+  gdouble major_step, minor_step;
+  gint nmajor = 0;
+  gint nminor = 0;
 
   ticks = &data->gradient;
 
@@ -2288,8 +2794,11 @@ gtk_plot_data_calc_gradient(GtkPlotData *data)
 
   nmajor = 0;
   ticks->step = major_step = (absmax - absmin) / (ticks->nmajorticks);
+  minor_step = major_step / ((gdouble)ticks->nminor + 1.0);
+
 
   ticks->major_values = (gdouble *)g_malloc((ticks->nmajorticks + 1) * sizeof(gdouble));
+  ticks->minor_values = (gdouble *)g_malloc((ticks->nmajorticks * ticks->nminorticks + 1) * sizeof(gdouble));
 
   if(ticks->step > 0.){
    tick = min - major_step;
@@ -2300,7 +2809,102 @@ gtk_plot_data_calc_gradient(GtkPlotData *data)
      }
      tick += major_step;
    }
+   if(ticks->nminor > 0){
+     gint n;
+     for(nmajor = 0; nmajor <= ticks->nmajorticks; nmajor++){
+        if(nmajor < ticks->nmajorticks)
+          tick = ticks->major_values[nmajor] - major_step;
+        else
+          tick = ticks->major_values[nmajor-1];
+        for(n = 1; n <= ticks->nminor; n++){
+          tick += minor_step;
+          if(tick >= absmin-1.E-10 && tick <= absmax+1.E-10){
+            nminor++;
+            ticks->minor_values[nminor-1] = tick;
+            ticks->nminorticks = nminor;
+          }
+        }
+     }  
+
+   }
   }
 
 }
 
+GtkPlotMarker *
+gtk_plot_data_add_marker(GtkPlotData *data, guint point)
+{
+  GtkPlotMarker *marker = NULL;
+/*
+  GList *list = NULL;
+*/
+
+  if(point >= data->num_points) return FALSE;
+
+/*
+  list = data->markers;
+  while(list){
+    marker = (GtkPlotMarker *)list->data;
+    if(marker->point == point) return FALSE;
+    list = list->next;
+  }
+*/
+
+  marker = g_new0(GtkPlotMarker, 1);
+
+  marker->data = data;
+  marker->point = point;  
+
+  data->markers = g_list_append(data->markers, marker);
+  return (marker);
+}
+
+gboolean
+gtk_plot_data_remove_marker(GtkPlotData *data, GtkPlotMarker *marker)
+{
+  GList *list;
+
+  list = data->markers;
+  while(list){
+    GtkPlotMarker *point;
+
+    point = (GtkPlotMarker *)list->data;
+    if(point == marker){
+      g_free(marker);
+      data->markers = g_list_remove_link(data->markers, list);
+      g_list_free_1(list);
+      return TRUE;
+    }
+    list = list->next;
+  } 
+
+  return FALSE;
+} 
+
+void
+gtk_plot_data_remove_markers(GtkPlotData *data)
+{
+  GList *list;
+
+  list = data->markers;
+  while(list){
+    g_free(list->data);
+    data->markers = g_list_remove_link(data->markers, list); 
+    g_list_free_1(list);
+    list = data->markers; 
+  }
+
+  data->markers = NULL;
+}
+
+void
+gtk_plot_data_show_markers(GtkPlotData *data, gboolean show)
+{
+  data->show_markers = show;
+}
+
+gboolean
+gtk_plot_data_markers_visible(GtkPlotData *data)
+{
+  return (data->show_markers);
+}
