@@ -35,6 +35,7 @@
 #include <time.h>
 #include <gtk/gtk.h>
 #include <wctype.h>
+#include <locale.h>
 
 #include "gtkplot.h"
 #include "gtkpsfont.h"
@@ -47,6 +48,8 @@ static void gtk_plot_ps_finalize 		(GtkObject *object);
 /*********************************************************************/
 /* Postscript specific functions */
 static gboolean psinit				(GtkPlotPC *pc); 
+static void pssetviewport			(GtkPlotPC *pc, 
+						 gdouble w, gdouble h); 
 static void psleave				(GtkPlotPC *pc);
 static void psgsave				(GtkPlotPC *pc);
 static void psgrestore				(GtkPlotPC *pc);
@@ -109,16 +112,19 @@ static void pssetdash				(GtkPlotPC *pc,
 						 gint num_values);
 static void psdrawpixmap                        (GtkPlotPC *pc,
                                                  GdkPixmap *pixmap,
+                                                 GdkBitmap *mask,
                                                  gint xsrc, gint ysrc,
                                                  gint xdest, gint ydest,
                                                  gint width, gint height,
                                                  gdouble sx, gdouble sy);
 
+static void ps_reencode_font			(FILE *file, char *fontname);
 static void color_to_hex                        (GdkColor color,
                                                  gchar string[7]);
 
 /*********************************************************************/
 static GtkPlotPCClass *parent_class = NULL;
+static gchar *locale = NULL;
 
 GtkType
 gtk_plot_ps_get_type (void)
@@ -165,6 +171,7 @@ gtk_plot_ps_class_init (GtkPlotPSClass *klass)
 
   pc_class->init = psinit;
   pc_class->leave = psleave;
+  pc_class->set_viewport = pssetviewport;
   pc_class->gsave = psgsave;
   pc_class->grestore = psgrestore;
   pc_class->clip = psclip;
@@ -319,6 +326,10 @@ gtk_plot_ps_set_size                            (GtkPlotPS *ps,
         ps->page_height = height;
    }
 
+   if(ps->orientation == GTK_PLOT_PORTRAIT)
+     gtk_plot_pc_set_viewport(GTK_PLOT_PC(ps), ps->page_width, ps->page_height);
+   else
+     gtk_plot_pc_set_viewport(GTK_PLOT_PC(ps), ps->page_height, ps->page_width);
 }
 
 void
@@ -330,6 +341,12 @@ gtk_plot_ps_set_scale                           (GtkPlotPS *ps,
   ps->scaley = scaley; 
 }
 
+static void pssetviewport			(GtkPlotPC *pc,
+						 gdouble w, gdouble h) 
+{
+
+}
+
 static void pssetlineattr			(GtkPlotPC *pc, 
                                                  gfloat line_width,
                                                  GdkLineStyle line_style,
@@ -338,7 +355,7 @@ static void pssetlineattr			(GtkPlotPC *pc,
 {
     FILE *psout = GTK_PLOT_PS(pc)->psfile;
 
-    fprintf(psout, "%f slw\n", line_width);
+    fprintf(psout, "%g slw\n", line_width);
     fprintf(psout, "%d slc\n", abs(cap_style - 1));
     fprintf(psout, "%d slj\n", join_style);
 
@@ -384,6 +401,8 @@ psleave(GtkPlotPC *pc)
     fprintf(GTK_PLOT_PS(pc)->psfile, "%%%%Trailer\n");
     fprintf(GTK_PLOT_PS(pc)->psfile, "%%%%EOF\n");
     fclose(GTK_PLOT_PS(pc)->psfile);
+    setlocale(LC_NUMERIC, locale);
+    g_free(locale);
 }
 
 static gboolean 
@@ -395,6 +414,9 @@ psinit						(GtkPlotPC *pc)
 
     now = time(NULL);
 
+    locale = g_strdup(setlocale(LC_NUMERIC, NULL));
+    setlocale(LC_NUMERIC, "C");
+
     ps = GTK_PLOT_PS(pc);
     psout = ps->psfile;
 
@@ -490,118 +512,78 @@ psinit						(GtkPlotPC *pc)
              "   end\n"
              "} def\n\n"
     ); 
-    
-   /*
-    * CREATE ISOLatin1Encoding if not
-    * there already; actually, this differs slightly from ISO, since ISO
-    * leaves /hyphen at pos. 8#055 (decimal 45), plus it has /hyphen at
-    * 8#255 Here, we put /minus at 8#055. Phew.
-    */
-
-    fprintf(psout, "\n"
-	"mark             \n"
-	"/ISOLatin1Encoding \n"
-        "8#000 1 8#054 {StandardEncoding exch get} for \n"
-        "/minus \n"
-        "8#056 1 8#217 {StandardEncoding exch get} for \n"
-        "/dotlessi \n"
-        "8#301 1 8#317 {StandardEncoding exch get} for \n"
-        "/space /exclamdown /cent /sterling /currency /yen /brokenbar /section \n"
-        "/dieresis /copyright /ordfeminine /guillemotleft /logicalnot /hyphen \n"
-        "/registered /macron /degree /plusminus /twosuperior /threesuperior /acute \n"
-        "/mu /paragraph /periodcentered /cedilla /onesuperior /ordmasculine \n"
-        "/guillemotright /onequarter /onehalf /threequarters /questiondown /Agrave \n"
-        "/Aacute /Acircumflex /Atilde /Adieresis /Aring /AE /Ccedilla /Egrave /Eacute\n"
-        "/Ecircumflex /Edieresis /Igrave /Iacute /Icircumflex /Idieresis /Eth /Ntilde\n"
-        "/Ograve /Oacute /Ocircumflex /Otilde /Odieresis /multiply /Oslash /Ugrave \n"
-        "/Uacute /Ucircumflex /Udieresis /Yacute /Thorn /germandbls /agrave /aacute \n"
-        "/acircumflex /atilde /adieresis /aring /ae /ccedilla /egrave /eacute \n"
-        "/ecircumflex /edieresis /igrave /iacute /icircumflex /idieresis /eth /ntilde\n"
-        "/ograve /oacute /ocircumflex /otilde /odieresis /divide /oslash /ugrave \n"
-        "/uacute /ucircumflex /udieresis /yacute /thorn /ydieresis \n"
-        "/ISOLatin1Encoding where not {256 array astore def} if \n"
-        "cleartomark \n");
-
-    /* Define makeISOEncoded procedure expecting 'fontname' and returning the
-     * font on stack; the procedure installs new font 'fontname-ISOLatin1' for
-     * later use.  this can be used as either
-     * /Times-Roman makeISOEncoded  pop ...
-     * /Times-Roman-ISOL1 findfont 55 scalefont setfont
-     * (when fonts are found many times) or
-     * /Times-Roman makeISOEncoded 55 scalefont setfont
-     * (for one-shot deals)
-     */
-
-    fprintf(psout, "\n"
-	"/makeISOEncoded \n"
-	"{ findfont /curfont exch def \n"
-        "  /newfont curfont maxlength dict def  \n"
-        "  /ISOLatin1 (-ISOLatin1) def\n"
-        "  /curfontname curfont /FontName get dup length string cvs def \n"
-        "  /newfontname curfontname length ISOLatin1 length add string \n"
-	" dup 0			curfontname putinterval \n"
-	" dup curfontname length ISOLatin1   putinterval \n"
-        "  def \n"
-        "  curfont   \n"
-        "  { exch dup /FID ne  \n"
-	"{ dup /Encoding eq  \n"
-        "  { exch pop ISOLatin1Encoding exch }  \n"
-        "  if  \n"
-        "  dup /FontName eq  \n"
-        "  { exch pop newfontname exch }	 \n"
-        "  if  \n"
-        "  exch newfont 3 1 roll put  \n"
-	"}  \n"
-	"{ pop pop }  \n"
-	"ifelse	\n"
-        "  }  \n"
-        "  forall \n"
-        "  newfontname newfont definefont \n"
-	"} def \n");
-
-    fprintf(psout, "\n"
-                   "/Times-Roman makeISOEncoded pop \n"
-                   "/Times-Italic makeISOEncoded pop \n"
-                   "/Times-Bold makeISOEncoded pop \n"
-                   "/Times-BoldItalic makeISOEncoded pop \n"
-                   "/AvantGarde-Book makeISOEncoded pop \n"
-                   "/AvantGarde-BookOblique makeISOEncoded pop \n"
-                   "/AvantGarde-Demi makeISOEncoded pop \n"
-                   "/AvantGarde-DemiOblique makeISOEncoded pop \n"
-                   "/Bookman-Light makeISOEncoded pop \n"
-                   "/Bookman-LightItalic makeISOEncoded pop \n"
-                   "/Bookman-Demi makeISOEncoded pop \n"
-                   "/Bookman-DemiItalic makeISOEncoded pop \n"
-                   "/Courier makeISOEncoded pop \n"
-                   "/Courier-Oblique makeISOEncoded pop \n"
-                   "/Courier-Bold makeISOEncoded pop \n"
-                   "/Courier-BoldOblique makeISOEncoded pop \n"
-                   "/Helvetica makeISOEncoded pop \n"
-                   "/Helvetica-Bold makeISOEncoded pop \n"
-                   "/Helvetica-Oblique makeISOEncoded pop \n"
-                   "/Helvetica-BoldOblique makeISOEncoded pop \n"
-                   "/Helvetica-Narrow makeISOEncoded pop \n"
-                   "/Helvetica-Narrow-Bold makeISOEncoded pop \n"
-                   "/Helvetica-Narrow-Oblique makeISOEncoded pop \n"
-                   "/Helvetica-Narrow-BoldOblique makeISOEncoded pop \n"
-                   "/NewCenturySchoolbook-Roman makeISOEncoded pop \n"
-                   "/NewCenturySchoolbook-Italic makeISOEncoded pop \n"
-                   "/NewCenturySchoolbook-Bold makeISOEncoded pop \n"
-                   "/NewCenturySchoolbook-BoldItalic makeISOEncoded pop \n"
-                   "/Palatino-Roman makeISOEncoded pop \n"
-                   "/Palatino-Italic makeISOEncoded pop \n"
-                   "/Palatino-Bold makeISOEncoded pop \n"
-                   "/Palatino-BoldItalic makeISOEncoded pop \n"
-    );
-
+  
+    fprintf(psout,
+          "[ /.notdef /.notdef /.notdef /.notdef /.notdef /.notdef /.notdef /.notdef /.notdef /.notdef\n"
+          "/.notdef /.notdef /.notdef /.notdef /.notdef /.notdef /.notdef /.notdef /.notdef /.notdef\n"
+          "/.notdef /.notdef /.notdef /.notdef /.notdef /.notdef /.notdef /.notdef /.notdef /.notdef\n"
+          "/.notdef /.notdef /space /exclam /quotedbl /numbersign /dollar /percent /ampersand /quoteright\n"
+          "/parenleft /parenright /asterisk /plus /comma /hyphen /period /slash /zero /one\n"
+          "/two /three /four /five /six /seven /eight /nine /colon /semicolon\n"          "/less /equal /greater /question /at /A /B /C /D /E\n"
+          "/F /G /H /I /J /K /L /M /N /O\n"
+          "/P /Q /R /S /T /U /V /W /X /Y\n"
+          "/Z /bracketleft /backslash /bracketright /asciicircum /underscore /quoteleft /a /b /c\n"
+          "/d /e /f /g /h /i /j /k /l /m\n"
+          "/n /o /p /q /r /s /t /u /v /w\n"
+          "/x /y /z /braceleft /bar /braceright /asciitilde /.notdef /.notdef /.notdef\n"
+          "/.notdef /.notdef /.notdef /.notdef /.notdef /.notdef /.notdef /.notdef /.notdef /.notdef\n"
+          "/.notdef /.notdef /.notdef /.notdef /.notdef /.notdef /.notdef /.notdef /.notdef /.notdef\n"
+          "/.notdef /.notdef /.notdef /.notdef /.notdef /.notdef /.notdef /.notdef /.notdef /.notdef\n"
+          "/space /exclamdown /cent /sterling /currency /yen /brokenbar /section /dieresis /copyright\n"
+          "/ordfeminine /guillemotleft /logicalnot /hyphen /registered /macron /degree /plusminus /twosuperior /threesuperior\n"
+          "/acute /mu /paragraph /periodcentered /cedilla /onesuperior /ordmasculine /guillemotright /onequarter /onehalf\n"
+          "/threequarters /questiondown /Agrave /Aacute /Acircumflex /Atilde /Adieresis /Aring /AE /Ccedilla\n"
+          "/Egrave /Eacute /Ecircumflex /Edieresis /Igrave /Iacute /Icircumflex /Idieresis /Eth /Ntilde\n"
+          "/Ograve /Oacute /Ocircumflex /Otilde /Odieresis /multiply /Oslash /Ugrave /Uacute /Ucircumflex\n"
+          "/Udieresis /Yacute /Thorn /germandbls /agrave /aacute /acircumflex /atilde /adieresis /aring\n"
+          "/ae /ccedilla /egrave /eacute /ecircumflex /edieresis /igrave /iacute /icircumflex /idieresis\n"
+          "/eth /ntilde /ograve /oacute /ocircumflex /otilde /odieresis /divide /oslash /ugrave\n"
+          "/uacute /ucircumflex /udieresis /yacute /thorn /ydieresis] /isolatin1encoding exch def\n");
+ 
+    ps_reencode_font(psout, "Times-Roman");
+    ps_reencode_font(psout, "Times-Italic");
+    ps_reencode_font(psout, "Times-Bold");
+    ps_reencode_font(psout, "Times-BoldItalic");
+    ps_reencode_font(psout, "AvantGarde-Book");
+    ps_reencode_font(psout, "AvantGarde-BookOblique");
+    ps_reencode_font(psout, "AvantGarde-Demi");
+    ps_reencode_font(psout, "AvantGarde-DemiOblique");
+    ps_reencode_font(psout, "Bookman-Light");
+    ps_reencode_font(psout, "Bookman-LightItalic");
+    ps_reencode_font(psout, "Bookman-Demi");
+    ps_reencode_font(psout, "Bookman-DemiItalic");
+    ps_reencode_font(psout, "Courier");
+    ps_reencode_font(psout, "Courier-Oblique");
+    ps_reencode_font(psout, "Courier-Bold");
+    ps_reencode_font(psout, "Courier-BoldOblique");
+    ps_reencode_font(psout, "Helvetica");
+    ps_reencode_font(psout, "Helvetica-Oblique");
+    ps_reencode_font(psout, "Helvetica-Bold");
+    ps_reencode_font(psout, "Helvetica-BoldOblique");
+    ps_reencode_font(psout, "Helvetica-Narrow");
+    ps_reencode_font(psout, "Helvetica-Narrow-Oblique");
+    ps_reencode_font(psout, "Helvetica-Narrow-Bold");
+    ps_reencode_font(psout, "Helvetica-Narrow-BoldOblique");
+    ps_reencode_font(psout, "NewCenturySchoolbook-Roman");
+    ps_reencode_font(psout, "NewCenturySchoolbook-Italic");
+    ps_reencode_font(psout, "NewCenturySchoolbook-Bold");
+    ps_reencode_font(psout, "NewCenturySchoolbook-BoldItalic");
+    ps_reencode_font(psout, "Palatino-Roman");
+    ps_reencode_font(psout, "Palatino-Italic");
+    ps_reencode_font(psout, "Palatino-Bold");
+    ps_reencode_font(psout, "Palatino-BoldItalic");
+    ps_reencode_font(psout, "Symbol");
+    ps_reencode_font(psout, "ZapfChancery-MediumItalic");
+    ps_reencode_font(psout, "ZapfDingbats");
+   
     if(ps->orientation == GTK_PLOT_PORTRAIT)
              fprintf(psout, "%d %d translate\n"
-                            "%f %f scale\n",
+                            "%g %g scale\n",
                             0, ps->page_height,
                             ps->scalex, -ps->scaley);
 
     if(ps->orientation == GTK_PLOT_LANDSCAPE)
-             fprintf(psout, "%f %f scale\n"
+             fprintf(psout, "%g %g scale\n"
                             "-90 rotate \n",
                             ps->scalex, -ps->scaley);
 
@@ -610,11 +592,31 @@ psinit						(GtkPlotPC *pc)
     return TRUE;
 }
 
+static void ps_reencode_font(FILE *file, char *fontname)
+{
+  /* Don't reencode the Symbol font, as it doesn't work in latin1 encoding.
+   * Instead, just define Symbol-latin1 to be the same as Symbol. */
+  if (!strcmp(fontname, "Symbol"))
+    fprintf(file,
+            "/%s-latin1\n"
+            "    /%s findfont\n"
+            "definefont pop\n", fontname, fontname);
+  else
+    fprintf(file,
+            "/%s-latin1\n"
+            "    /%s findfont\n"
+            "    dup length dict begin\n"
+            "   {1 index /FID ne {def} {pop pop} ifelse} forall\n"
+            "   /Encoding isolatin1encoding def\n"
+            "    currentdict end\n"
+            "definefont pop\n", fontname, fontname);
+}
+
 static void pssetcolor(GtkPlotPC *pc, const GdkColor *color)
 {
     FILE *psout = GTK_PLOT_PS(pc)->psfile;
 
-    fprintf(psout, "%f %f %f setrgbcolor\n",
+    fprintf(psout, "%g %g %g setrgbcolor\n",
 	    (gdouble) color->red / 65535.0,
 	    (gdouble) color->green / 65535.0,
 	    (gdouble) color->blue / 65535.0);
@@ -625,8 +627,9 @@ psdrawpoint(GtkPlotPC *pc, gdouble x, gdouble y)
 {
   FILE *psout = GTK_PLOT_PS(pc)->psfile;
 
-  fprintf(psout, "%f %f m\n", x, y);
-  fprintf(psout, "%f %f l\n", x, y);
+  fprintf(psout, "n\n");
+  fprintf(psout, "%g %g m\n", x, y);
+  fprintf(psout, "%g %g l\n", x, y);
   fprintf(psout, "s\n");
 }
 
@@ -637,9 +640,9 @@ psdrawlines(GtkPlotPC *pc, GtkPlotPoint *points, gint numpoints)
   FILE *psout = GTK_PLOT_PS(pc)->psfile;
  
   fprintf(psout,"n\n");
-  fprintf(psout,"%f %f m\n", points[0].x, points[0].y);
+  fprintf(psout,"%g %g m\n", points[0].x, points[0].y);
   for(i = 1; i < numpoints; i++)
-        fprintf(psout,"%f %f l\n", points[i].x, points[i].y);
+        fprintf(psout,"%g %g l\n", points[i].x, points[i].y);
 
   fprintf(psout,"s\n");
 }
@@ -651,9 +654,9 @@ psdrawpolygon(GtkPlotPC *pc, gboolean filled, GtkPlotPoint *points, gint numpoin
   FILE *psout = GTK_PLOT_PS(pc)->psfile;
 
   fprintf(psout,"n\n");
-  fprintf(psout,"%f %f m\n", points[0].x, points[0].y);
+  fprintf(psout,"%g %g m\n", points[0].x, points[0].y);
   for(i = 1; i < numpoints; i++)
-      fprintf(psout,"%f %f l\n", points[i].x, points[i].y);
+      fprintf(psout,"%g %g l\n", points[i].x, points[i].y);
 
   if(filled)
      fprintf(psout,"f\n");
@@ -667,8 +670,8 @@ static void psdrawline(GtkPlotPC *pc, gdouble x0, gdouble y0, gdouble xf, gdoubl
 {
   FILE *psout = GTK_PLOT_PS(pc)->psfile;
 
-  fprintf(psout, "%f %f m\n", x0, y0);
-  fprintf(psout, "%f %f l\n", xf, yf);
+  fprintf(psout, "%g %g m\n", x0, y0);
+  fprintf(psout, "%g %g l\n", xf, yf);
   fprintf(psout, "s\n");
 }
 
@@ -695,7 +698,7 @@ psdrawcircle(GtkPlotPC *pc, gboolean filled, gdouble x, gdouble y, gdouble size)
 {
   FILE *psout = GTK_PLOT_PS(pc)->psfile;
 
-  fprintf(psout,"n %f %f %f %f 0 360 ellipse\n", 
+  fprintf(psout,"n %g %g %g %g 0 360 ellipse\n", 
           x, y, size / 2., size / 2.);
 
   if(filled)
@@ -712,7 +715,7 @@ psdrawellipse(GtkPlotPC *pc,
 {
   FILE *psout = GTK_PLOT_PS(pc)->psfile;
 
-  fprintf(psout,"n %f %f %f %f 0 360 ellipse\n", 
+  fprintf(psout,"n %g %g %g %g 0 360 ellipse\n", 
           x+width/2., y+height/2., 
           width/2., height/2.);
 
@@ -1141,7 +1144,8 @@ pssetfont(GtkPlotPC *pc, GtkPSFont *psfont, gint height)
 	    "/%s ff [0 1 -1 0 0 0.3] makefont [%d 0 0 %d 0 0] makefont sf\n",
 	    psfont->psname, height, height);
   else
-    fprintf(psout, "/%s ff %f scf sf\n", psfont->psname, (double)height);
+    fprintf(psout, "/%s-latin1 ff %g scf sf\n", psfont->psname, (double)height);
+
 }
 
 
@@ -1240,9 +1244,9 @@ psclipmask(GtkPlotPC *pc, gdouble x, gdouble y, const GdkBitmap *mask)
   fprintf(psout,"gsave\n");
 
   fprintf(psout,"n\n");
-  fprintf(psout,"%f %f m\n", x + points[0].x, y + points[0].y);
+  fprintf(psout,"%g %g m\n", x + points[0].x, y + points[0].y);
   for(i = 1; i < npoints; i++)
-      fprintf(psout,"%f %f l\n", x + points[i].x, y + points[i].y);
+      fprintf(psout,"%g %g l\n", x + points[i].x, y + points[i].y);
 
   fprintf(psout,"cp\n");
 
@@ -1270,6 +1274,7 @@ psclip(GtkPlotPC *pc, const GdkRectangle *clip)
 static void 
 psdrawpixmap  (GtkPlotPC *pc,
                GdkPixmap *pixmap,
+               GdkBitmap *mask,
                gint xsrc, gint ysrc,
                gint xdest, gint ydest,
                gint width, gint height,
@@ -1293,14 +1298,16 @@ psdrawpixmap  (GtkPlotPC *pc,
                           xsrc, ysrc,
                           width, height);
 
+    if(mask) gtk_plot_pc_clip_mask(pc, xdest, ydest, mask);
 
-    fprintf(psout, "%d %f translate\n", xdest, ydest + height * scale_y);
-    fprintf(psout, "%f %f scale\n",width * scale_x, height * scale_y);
+    fprintf(psout, "%d %g translate\n", xdest, ydest + height * scale_y);
+    fprintf(psout, "%g %g scale\n",width * scale_x, height * scale_y);
     fprintf(psout, "%d %d 8 [%d 0 0 %d 0 %d]\n",width, height, width, height, height);
     fprintf(psout, "/scanline %d 3 mul string def\n", width);
     fprintf(psout, "{ currentfile scanline readhexstring pop } false 3\n");
     fprintf(psout, "colorimage\n");
 
+
     for(y = 0; y < height; y++){
       for(x = 0; x < width; x++){
         GdkColor color;
@@ -1317,6 +1324,7 @@ psdrawpixmap  (GtkPlotPC *pc,
     }
 
     gdk_image_destroy(image);
+    if(mask) gtk_plot_pc_clip_mask(pc, xdest, ydest, NULL);
   }
 
   gdk_color_context_free(cc);
