@@ -32,6 +32,13 @@
 
 static void gtk_plot_surface_class_init 	(GtkPlotSurfaceClass *klass);
 static void gtk_plot_surface_init 		(GtkPlotSurface *data);
+static void gtk_plot_surface_destroy 		(GtkObject *object);
+static void gtk_plot_surface_update_range 	(GtkPlotData *data);
+static void gtk_plot_surface_real_build_mesh	(GtkPlotSurface *surface);
+static void update_data 			(GtkPlotData *data, 
+						 gboolean new_range);
+static gboolean gtk_plot_surface_add_to_plot 	(GtkPlotData *data,
+						 GtkPlot *plot);
 static void gtk_plot_surface_draw 		(GtkWidget *widget, 
 						 GdkRectangle *area);
 static void gtk_plot_surface_draw_private 	(GtkPlotData *data);
@@ -39,7 +46,7 @@ static void gtk_plot_surface_draw_legend	(GtkPlotData *data,
 						 gint x, gint y);
 static void gtk_plot_surface_real_draw		(GtkPlotSurface *data); 
 static void gtk_plot_surface_draw_triangle 	(GtkPlotSurface *surface,
-                                		 GtkPlotVector *points,
+                                		 GtkPlotDTtriangle *triangle,
                                                  gint sign);
 static void gtk_plot_surface_lighting 		(GdkColor *a, 
 						 GdkColor *b, 
@@ -58,6 +65,8 @@ static void rgb_to_hsv 				(gdouble  r,
             					 gdouble *h, 
 						 gdouble *s, 
 						 gdouble *v);
+static gint compare_func                           (gpointer a, gpointer b);
+
 
 static GtkDataClass *parent_class = NULL;
 
@@ -100,127 +109,58 @@ gtk_plot_surface_class_init (GtkPlotSurfaceClass *klass)
   data_class = (GtkPlotDataClass *) klass;
   surface_class = (GtkPlotSurfaceClass *) klass;
 
+  object_class->destroy = gtk_plot_surface_destroy;
+
   widget_class->draw = gtk_plot_surface_draw;
 
+  data_class->update = update_data;
+  data_class->add_to_plot = gtk_plot_surface_add_to_plot;
   data_class->draw_data = gtk_plot_surface_draw_private;
   data_class->draw_legend = gtk_plot_surface_draw_legend;
   surface_class->draw_triangle = gtk_plot_surface_draw_triangle;
 
 }
 
-
 static void
-gtk_plot_surface_init (GtkPlotSurface *dataset)
+update_data (GtkPlotData *data, gboolean new_range)
 {
-  GtkWidget *widget;
-  GdkColormap *colormap;
-  GdkColor color;
+  if(new_range && data->is_function) 
+    GTK_PLOT_SURFACE(data)->recalc_dt = TRUE;
 
-  GTK_WIDGET_SET_FLAGS(dataset, GTK_NO_WINDOW);
-
-  widget = GTK_WIDGET(dataset);
-  colormap = gtk_widget_get_colormap(widget);
-
-  gdk_color_parse("black", &color);
-  gdk_color_alloc(colormap, &color);
-  dataset->grid_foreground = color;
-
-  gdk_color_parse("dark green", &color);
-  gdk_color_alloc(colormap, &color);
-  dataset->grid_background = color;
-
-  gdk_color_parse("gray30", &color);
-  gdk_color_alloc(colormap, &color);
-  dataset->shadow = color;
-
-  gdk_color_parse("blue", &color);
-  gdk_color_alloc(colormap, &color);
-  dataset->color = color;
-
-  dataset->light.x = 0.;
-  dataset->light.y = 0.;
-  dataset->light.z = 1.;
-
-  dataset->nx = 0;
-  dataset->ny = 0;
-
-  dataset->show_grid = TRUE;
-  dataset->show_mesh = FALSE;
-
-  dataset->ambient = 0.3;
-
-  dataset->xstep = 0.05;
-  dataset->ystep = 0.05;
-}
-
-GtkWidget*
-gtk_plot_surface_new (void)
-{
-  GtkPlotData *data;
-
-  data = gtk_type_new (gtk_plot_surface_get_type ());
-
-  return GTK_WIDGET (data);
-}
-
-GtkWidget*
-gtk_plot_surface_new_function (GtkPlotFunc3D function)
-{
-  GtkWidget *dataset;
-
-  dataset = gtk_type_new (gtk_plot_surface_get_type ());
-
-  gtk_plot_surface_construct_function(GTK_PLOT_SURFACE(dataset), function);
-
-  return dataset;
-}
-
-void
-gtk_plot_surface_construct_function(GtkPlotSurface *surface, 
-				   GtkPlotFunc3D function)
-{
-  GTK_PLOT_DATA(surface)->is_function = TRUE;
-  GTK_PLOT_DATA(surface)->function3d = function;
+  GTK_PLOT_DATA_CLASS(parent_class)->update(data, new_range);
 }
 
 static void
-gtk_plot_surface_draw (GtkWidget *widget, GdkRectangle *area)
+gtk_plot_surface_update_range (GtkPlotData *data)
 {
-  if(!GTK_WIDGET_VISIBLE(widget)) return;
-
-  gtk_plot_surface_draw_private(GTK_PLOT_DATA(widget));
-}
-
-static void
-gtk_plot_surface_draw_private   (GtkPlotData *data)  
-{
-  GtkPlot *plot;
   GtkPlotSurface *surface;
-  GtkPlotAxis *ax, *ay;
+  GtkPlot *plot;
 
-  g_return_if_fail(GTK_IS_PLOT_SURFACE(data));
+  if(!data->redraw_pending) return; 
 
   surface = GTK_PLOT_SURFACE(data);
-  data = GTK_PLOT_DATA(surface);
+  plot = data->plot;
 
-  g_return_if_fail(GTK_PLOT_DATA(data)->plot != NULL);
-  g_return_if_fail(GTK_IS_PLOT(GTK_PLOT_DATA(data)->plot));
-  g_return_if_fail(GTK_WIDGET_REALIZED(GTK_PLOT_DATA(data)->plot));
 
-  plot = GTK_PLOT(data->plot);
-
-  if(GTK_IS_PLOT3D(plot)){
-     ax = GTK_PLOT3D(plot)->ax;
-     ay = GTK_PLOT3D(plot)->ay;
-  }else{
-     ax = plot->bottom;
-     ay = plot->left;
+  if(!surface->recalc_dt && data->num_points > 0){
+     gtk_plot_surface_recalc_nodes(surface);
+     data->redraw_pending = FALSE;
+     surface->recalc_dt = FALSE;
+     return;
   }
 
-  if(!data->is_function)
-     gtk_plot_surface_real_draw(surface);
-  else
-   {
+  if(!data->is_function && data->num_points > 0){
+     if(surface->recalc_dt)
+       gtk_plot_surface_build_mesh(surface);
+     else
+       gtk_plot_surface_recalc_nodes(surface);
+
+     data->redraw_pending = FALSE;
+     surface->recalc_dt = FALSE;
+     return;
+  }
+
+  if(data->is_function){
      gdouble xstep, ystep;
      gdouble xmin, xmax, ymin, ymax;
      gdouble x, y;
@@ -269,15 +209,156 @@ gtk_plot_surface_draw_private   (GtkPlotData *data)
      data->y = fy;
      data->z = fz;
 
-     gtk_plot_surface_real_draw(surface);
+     data->num_points = npoints;
+     gtk_plot_surface_build_mesh(surface);
 
      g_free(fx);
      g_free(fy);
      g_free(fz);
 
      data->x = old_x; data->y = old_y; data->z = old_z;
-   }
+  }
 
+  data->redraw_pending = FALSE;
+  surface->recalc_dt = FALSE;
+}
+
+static gboolean 
+gtk_plot_surface_add_to_plot (GtkPlotData *data, GtkPlot *plot)
+{
+  if(!data->is_function && data->num_points > 0){
+    gtk_plot_surface_build_mesh(GTK_PLOT_SURFACE(data));
+  }
+
+  data->redraw_pending = TRUE;
+
+  return TRUE;
+}
+
+static void
+gtk_plot_surface_init (GtkPlotSurface *dataset)
+{
+  GtkWidget *widget;
+  GdkColormap *colormap;
+  GdkColor color;
+
+  GTK_WIDGET_SET_FLAGS(dataset, GTK_NO_WINDOW);
+
+  widget = GTK_WIDGET(dataset);
+  colormap = gtk_widget_get_colormap(widget);
+
+  gdk_color_parse("black", &color);
+  gdk_color_alloc(colormap, &color);
+  dataset->grid_foreground = color;
+
+  gdk_color_parse("dark green", &color);
+  gdk_color_alloc(colormap, &color);
+  dataset->grid_background = color;
+
+  gdk_color_parse("gray30", &color);
+  gdk_color_alloc(colormap, &color);
+  dataset->shadow = color;
+
+  gdk_color_parse("blue", &color);
+  gdk_color_alloc(colormap, &color);
+  dataset->color = color;
+
+  dataset->light.x = 0.;
+  dataset->light.y = 0.;
+  dataset->light.z = 1.;
+
+  dataset->nx = 0;
+  dataset->ny = 0;
+
+  dataset->show_grid = TRUE;
+  dataset->show_mesh = FALSE;
+  dataset->transparent = FALSE;
+  dataset->height_gradient = FALSE;
+
+  dataset->ambient = 0.3;
+
+  dataset->xstep = 0.05;
+  dataset->ystep = 0.05;
+
+  dataset->mesh_line.color = dataset->grid_foreground;
+  dataset->mesh_line.line_width = 1;
+  dataset->mesh_line.line_style = GTK_PLOT_LINE_SOLID;
+
+  dataset->dt = GTK_PLOT_DT(gtk_plot_dt_new(0));
+  dataset->recalc_dt = TRUE;
+}
+
+GtkWidget*
+gtk_plot_surface_new (void)
+{
+  GtkPlotData *data;
+
+  data = gtk_type_new (gtk_plot_surface_get_type ());
+
+  return GTK_WIDGET (data);
+}
+
+GtkWidget*
+gtk_plot_surface_new_function (GtkPlotFunc3D function)
+{
+  GtkWidget *dataset;
+
+  dataset = gtk_type_new (gtk_plot_surface_get_type ());
+
+  gtk_plot_surface_construct_function(GTK_PLOT_SURFACE(dataset), function);
+
+  return dataset;
+}
+
+void
+gtk_plot_surface_construct_function(GtkPlotSurface *surface, 
+				   GtkPlotFunc3D function)
+{
+  GTK_PLOT_DATA(surface)->is_function = TRUE;
+  GTK_PLOT_DATA(surface)->function3d = function;
+}
+
+void
+gtk_plot_surface_destroy(GtkObject *object) 
+{
+  GtkPlotSurface *surface;
+ 
+  surface = GTK_PLOT_SURFACE(object);
+
+  if(surface->dt)
+    gtk_object_destroy(GTK_OBJECT(surface->dt));
+
+  if ( GTK_OBJECT_CLASS (parent_class)->destroy )
+    (* GTK_OBJECT_CLASS (parent_class)->destroy) (object);
+}
+
+static void
+gtk_plot_surface_draw (GtkWidget *widget, GdkRectangle *area)
+{
+  if(!GTK_WIDGET_VISIBLE(widget)) return;
+
+  gtk_plot_surface_draw_private(GTK_PLOT_DATA(widget));
+}
+
+static void
+gtk_plot_surface_draw_private   (GtkPlotData *data)  
+{
+  GtkPlot *plot;
+  GtkPlotSurface *surface;
+
+  g_return_if_fail(GTK_IS_PLOT_SURFACE(data));
+
+  surface = GTK_PLOT_SURFACE(data);
+  data = GTK_PLOT_DATA(surface);
+
+  g_return_if_fail(GTK_PLOT_DATA(data)->plot != NULL);
+  g_return_if_fail(GTK_IS_PLOT(GTK_PLOT_DATA(data)->plot));
+
+  if(data->redraw_pending) gtk_plot_surface_update_range(data);
+
+  plot = GTK_PLOT(data->plot);
+
+  gtk_plot_surface_real_draw(surface);
 }
 
 static void
@@ -286,14 +367,7 @@ gtk_plot_surface_real_draw   (GtkPlotSurface *surface)
   GtkWidget *widget;
   GtkPlot *plot = NULL;
   GtkPlotData *dataset;
-  GdkRectangle area;
-  gint i;
-  gint nx, ny;
-  gdouble x1, y1, z1;
-  gdouble x2, y2, z2;
-  gint xa, xb, xstep;
-  gint ya, yb, ystep;
-  gdouble m;
+  GList *list;
 
   g_return_if_fail(GTK_IS_PLOT_SURFACE(surface));
 
@@ -301,7 +375,6 @@ gtk_plot_surface_real_draw   (GtkPlotSurface *surface)
 
   g_return_if_fail(GTK_PLOT_DATA(dataset)->plot != NULL);
   g_return_if_fail(GTK_IS_PLOT(GTK_PLOT_DATA(dataset)->plot));
-  g_return_if_fail(GTK_WIDGET_REALIZED(GTK_PLOT_DATA(dataset)->plot));
 
   plot = dataset->plot;
   widget = GTK_WIDGET(plot);
@@ -311,136 +384,53 @@ gtk_plot_surface_real_draw   (GtkPlotSurface *surface)
   if(!GTK_WIDGET_DRAWABLE(widget)) return;
   if(!GTK_WIDGET_VISIBLE(widget)) return;
 
-  m = plot->magnification;
+  gtk_plot_set_line_attributes(plot, surface->mesh_line);
 
-  area.x = widget->allocation.x;
-  area.y = widget->allocation.y;
-  area.width = widget->allocation.width;
-  area.height = widget->allocation.height;
-
-  xa = 0;
-  xb = surface->nx - 2;
-  xstep = 1;
-  ya = 0;
-  yb = surface->ny - 2;
-  ystep = 1;
-
-  gtk_plot_pc_set_lineattr(plot->pc, 1, 0, 0, 0);
-
-  if(GTK_IS_PLOT3D(plot)){
-    gtk_plot3d_get_pixel(GTK_PLOT3D(plot), 
-                         dataset->x[0],
-                         dataset->y[0],
-                         0.,
-                         &x1, &y1, &z1);
-    gtk_plot3d_get_pixel(GTK_PLOT3D(plot), 
-                         dataset->x[(surface->ny-1)*surface->nx],
-                         dataset->y[(surface->ny-1)*surface->nx],
-                         0.,
-                         &x2, &y2, &z2);
-
-    if(z2 > z1){
-      ya = surface->ny - 1;
-      yb = 1;
-      ystep = -1;
-    } 
-
-    gtk_plot3d_get_pixel(GTK_PLOT3D(plot), 
-                         dataset->x[surface->nx-1],
-                         dataset->y[surface->nx-1],
-                         0.,
-                         &x2, &y2, &z2);
-
-    if(z2 > z1){
-      xa = surface->nx - 1;
-      xb = 1;
-      xstep = -1;
-    } 
+  list = surface->dt->triangles;
+  while(list){
+    GtkPlotDTtriangle *t;
+ 
+    t = (GtkPlotDTtriangle *)list->data;
+    GTK_PLOT_SURFACE_CLASS(GTK_OBJECT(surface)->klass)->draw_triangle(surface, t, -1.); 
+    list = list->next;
   }
 
-  for(nx = xa; nx != xb + xstep; nx += xstep)
-    for(ny = ya; ny != yb + ystep; ny += ystep)
-    {
-      GtkPlotVector center;
-      GtkPlotVector v[4];
-      gint n[4];
-
-      n[0] = ny*surface->nx + nx;
-      n[1] = ny*surface->nx + nx + xstep;
-      n[2] = (ny + ystep)*surface->nx + nx + xstep;
-      n[3] = (ny + ystep)*surface->nx + nx;
-
-      center.x = center.y = center.z = 0.0;
-
-      for(i = 0; i < 4; i++){
-        v[i].x = dataset->x[n[i]];
-        v[i].y = dataset->y[n[i]];
-        v[i].z = dataset->z[n[i]];
-
-        center.x = center.x + v[i].x;
-        center.y = center.y + v[i].y;
-        center.z = center.z + v[i].z;
-      }
-
-      center.x /= 4.0;
-      center.y /= 4.0;
-      center.z /= 4.0;
-
-      for(i = 0; i < 4; i++){
-         GtkPlotVector point[3];
-         gint next;
-
-         next = (i == 3) ? 0 : i + 1;
-
-         point[0] = center;
-         point[1] = v[i];
-         point[2] = v[next];
-
-         GTK_PLOT_SURFACE_CLASS(GTK_OBJECT(surface)->klass)->draw_triangle(surface, point, -xstep*ystep); 
-      }
-    }
-
   gtk_plot_pc_grestore(plot->pc);
 }
 
 static void
 gtk_plot_surface_draw_triangle (GtkPlotSurface *surface,
-                                GtkPlotVector *points,
+                                GtkPlotDTtriangle *triangle,
                                 gint sign)
 {
   GtkPlot *plot;
   GtkPlotVector side1, side2, light, normal;
-  GdkDrawable *drawable;
+  GtkPlotVector points[3];
   GtkPlotPoint t[3];
-  GdkColor color;
-  gint i;
+  GdkDrawable *drawable;
+  GdkColor color, real_color;
   gdouble factor, norm;
   gboolean visible = TRUE;
 
   plot = GTK_PLOT(GTK_PLOT_DATA(surface)->plot);
   drawable = plot->drawable;
 
-  color = surface->color; 
+  points[0].x = triangle->na->x;
+  points[0].y = triangle->na->y;
+  points[0].z = triangle->na->z;
+  points[1].x = triangle->nb->x;
+  points[1].y = triangle->nb->y;
+  points[1].z = triangle->nb->z;
+  points[2].x = triangle->nc->x;
+  points[2].y = triangle->nc->y;
+  points[2].z = triangle->nc->z;
 
-  light = surface->light;
-  norm = sqrt(light.x*light.x + light.y*light.y + light.z*light.z); 
-  light.x /= norm;
-  light.y /= norm;
-  light.z /= norm;
-
-  for(i = 0; i < 3; i++){
-    gdouble x, y, z;
-    if(GTK_IS_PLOT3D(plot))
-      gtk_plot3d_get_pixel(GTK_PLOT3D(plot), 
-                           points[i].x, points[i].y, points[i].z,
-                           &x, &y, &z);
-    else
-      gtk_plot_get_pixel(plot, 
-                         points[i].x, points[i].y, &x, &y);
-
-    t[i].x = x;
-    t[i].y = y;
-  }
+  t[0].x = triangle->na->px;
+  t[0].y = triangle->na->py;
+  t[1].x = triangle->nb->px;
+  t[1].y = triangle->nb->py;
+  t[2].x = triangle->nc->px;
+  t[2].y = triangle->nc->py;
 
   side1.x = (points[1].x - points[0].x) * sign;
   side1.y = (points[1].y - points[0].y) * sign;
@@ -449,26 +439,45 @@ gtk_plot_surface_draw_triangle (GtkPlotSurface *surface,
   side2.y = (points[2].y - points[0].y) * sign;
   side2.z = (points[2].z - points[0].z) * sign;
          
-  normal.x = side1.y * side2.z - side1.z * side2.y;
-  normal.y = side1.z * side2.x - side1.x * side2.z;
-  normal.z = side1.x * side2.y - side1.y * side2.x;
+  if(surface->height_gradient){
+    gtk_plot_data_get_gradient_level(GTK_PLOT_DATA(surface),
+				     (triangle->na->z +
+                                      triangle->nb->z +
+				      triangle->nc->z) / 3.0,
+				     &real_color); 
+  }else{
+    color = surface->color; 
+    light = surface->light;
+    norm = sqrt(light.x*light.x + light.y*light.y + light.z*light.z); 
+    light.x /= norm;
+    light.y /= norm;
+    light.z /= norm;
 
-  norm = sqrt(normal.x*normal.x + normal.y*normal.y + normal.z*normal.z); 
-  factor = (normal.x*light.x + normal.y*light.y + normal.z*light.z) / norm;
+    if(GTK_IS_PLOT3D(plot)){
+      normal.x = side1.y * side2.z - side1.z * side2.y;
+      normal.y = side1.z * side2.x - side1.x * side2.z;
+      normal.z = side1.x * side2.y - side1.y * side2.x;
 
-  gtk_plot_surface_lighting(&surface->color, &color, 
-                            factor, surface->ambient); 
+      norm = sqrt(normal.x*normal.x + normal.y*normal.y + normal.z*normal.z); 
+      factor = (normal.x*light.x + normal.y*light.y + normal.z*light.z) / norm;
+    } else {
+      factor = 1.0;
+    }
+
+    gtk_plot_surface_lighting(&color, &real_color, 
+                              factor, surface->ambient); 
+  }
 
   if(GTK_IS_PLOT3D(plot))
     if(((t[1].x-t[0].x)*(t[2].y-t[0].y) - (t[1].y-t[0].y)*(t[2].x-t[0].x))*sign > 0)
          visible = FALSE;
 
   if(visible)
-         gtk_plot_pc_set_color(plot->pc, &color);
+         gtk_plot_pc_set_color(plot->pc, &real_color);
   else
          gtk_plot_pc_set_color(plot->pc, &surface->shadow);
 
-  gtk_plot_pc_draw_polygon(plot->pc, TRUE, t, 3); 
+  gtk_plot_pc_draw_polygon(plot->pc, !surface->transparent, t, 3); 
 
 
   if(visible)
@@ -476,10 +485,10 @@ gtk_plot_surface_draw_triangle (GtkPlotSurface *surface,
   else
          gtk_plot_pc_set_color(plot->pc, &surface->grid_background);
 
-  if(surface->show_mesh)
+  if(surface->show_mesh || !surface->dt->quadrilateral)
        gtk_plot_pc_draw_polygon(plot->pc, FALSE, t, 3); 
 
-  if(!surface->show_mesh && surface->show_grid)
+  if(!surface->show_mesh && surface->show_grid && surface->dt->quadrilateral)
        gtk_plot_pc_draw_line(plot->pc,  
                             t[1].x, t[1].y, t[2].x, t[2].y); 
                               
@@ -499,7 +508,6 @@ gtk_plot_surface_draw_legend(GtkPlotData *data, gint x, gint y)
 
   g_return_if_fail(data->plot != NULL);
   g_return_if_fail(GTK_IS_PLOT(data->plot));
-  g_return_if_fail(GTK_WIDGET_REALIZED(data->plot));
 
   plot = data->plot;
   area.x = GTK_WIDGET(plot)->allocation.x;
@@ -521,12 +529,29 @@ gtk_plot_surface_draw_legend(GtkPlotData *data, gint x, gint y)
                          &lascent, &ldescent);
 
 
-  gtk_plot_pc_set_color(plot->pc, &surface->color);
+  if(!surface->height_gradient){
+    gtk_plot_pc_set_color(plot->pc, &surface->color);
 
-  gtk_plot_pc_draw_rectangle(plot->pc, TRUE, 
-                             x, y,
-                             roundint(plot->legends_line_width * m), 
-                             lascent + ldescent);
+    gtk_plot_pc_draw_rectangle(plot->pc, TRUE, 
+                               x, y,
+                               roundint(plot->legends_line_width * m), 
+                               lascent + ldescent);
+  }else{
+    gdouble level, step;
+    gint lx = x, lstep;
+    step = (data->gradient.end - data->gradient.begin) / 10;
+    lstep = roundint(plot->legends_line_width * m / 10.);
+    for(level = data->gradient.begin; level < data->gradient.end; level += step){
+      GdkColor color;
+      gtk_plot_data_get_gradient_level(data, level, &color);
+      gtk_plot_pc_set_color(plot->pc, &color);
+
+      gtk_plot_pc_draw_rectangle(plot->pc, TRUE, 
+                                 lx, y,
+                                 lstep, lascent + ldescent);
+      lx += lstep;
+    }
+  }
 
 
   legend.x = (gdouble)(area.x + x + roundint((plot->legends_line_width + 4) * m))
@@ -534,6 +559,9 @@ gtk_plot_surface_draw_legend(GtkPlotData *data, gint x, gint y)
   legend.y = (gdouble)(area.y + y + lascent) / (gdouble)area.height;
 
   gtk_plot_draw_text(plot, legend);
+
+  y += 2*lheight;
+  gtk_plot_data_draw_gradient(data, x, y);
 }
 
 
@@ -555,6 +583,11 @@ gtk_plot_surface_lighting (GdkColor *a, GdkColor *b,
   gdouble red, green, blue;
   gdouble h, s, v;
 
+  if(normal == 1.0){
+   *b = *a;
+   return;
+  }
+
   normal = MIN(fabs(normal), 1.0);
 
   red = a->red;
@@ -766,6 +799,13 @@ gtk_plot_surface_set_light      (GtkPlotSurface *data,
   data->light.z = z;
 }
 
+void            
+gtk_plot_surface_use_height_gradient (GtkPlotSurface *data,
+                                   gboolean use_gradient)
+{
+  data->height_gradient = use_gradient;
+}
+
 void            
 gtk_plot_surface_set_ambient      (GtkPlotSurface *data,
                                    gdouble ambient)
@@ -773,6 +813,13 @@ gtk_plot_surface_set_ambient      (GtkPlotSurface *data,
   data->ambient = ambient;
 }
 
+void            
+gtk_plot_surface_set_transparent  (GtkPlotSurface *data,
+                                   gboolean transparent)
+{
+  data->transparent = transparent;
+}
+
 /******************************************
  * gtk_plot_surface_set_points
  * gtk_plot_surface_get_points
@@ -800,9 +847,9 @@ gtk_plot_surface_set_ambient      (GtkPlotSurface *data,
 
 void
 gtk_plot_surface_set_points(GtkPlotSurface *data, 
-                         gdouble *x, gdouble *y, gdouble *z,
-                         gdouble *dx, gdouble *dy, gdouble *dz,
-                         gint nx, gint ny)
+                            gdouble *x, gdouble *y, gdouble *z,
+                            gdouble *dx, gdouble *dy, gdouble *dz,
+                            gint nx, gint ny)
 {
   GTK_PLOT_DATA(data)->x = x;
   GTK_PLOT_DATA(data)->y = y;
@@ -812,6 +859,9 @@ gtk_plot_surface_set_points(GtkPlotSurface *data,
   GTK_PLOT_DATA(data)->dz = dz;
   data->nx = nx;
   data->ny = ny;
+  GTK_PLOT_DATA(data)->num_points = nx * ny;
+
+  gtk_plot_surface_build_mesh(data);
 }
 
 void
@@ -960,3 +1010,188 @@ gtk_plot_surface_get_ystep(GtkPlotSurface *dataset)
 {
   return (dataset->ystep);
 }
+
+void
+gtk_plot_surface_build_mesh(GtkPlotSurface *surface)
+{
+  GtkPlotData *data;
+  GtkPlot *plot;
+
+  data = GTK_PLOT_DATA(surface);
+  if(!data->plot) return;
+  plot = data->plot;
+
+  if(data->is_function){
+     gdouble xstep, ystep;
+     gdouble xmin, xmax, ymin, ymax;
+     gdouble x, y;
+     gdouble *fx = NULL, *fy = NULL, *fz = NULL;
+     gdouble *old_x, *old_y, *old_z;
+     gint nx, ny;
+     gint npoints;
+
+     xmin = GTK_PLOT(plot)->xmin;
+     xmax = GTK_PLOT(plot)->xmax;
+     ymin = GTK_PLOT(plot)->ymin;
+     ymax = GTK_PLOT(plot)->ymax;
+
+     xstep = surface->xstep;
+     surface->nx = roundint((xmax - xmin) / xstep) + 1;
+
+     ystep = surface->ystep;
+     surface->ny = roundint((ymax - ymin) / ystep) + 1;
+
+     npoints = surface->nx * surface->ny;
+     fx = (gdouble *)g_malloc((npoints + 1) * sizeof(gdouble));
+     fy = (gdouble *)g_malloc((npoints + 1) * sizeof(gdouble));
+     fz = (gdouble *)g_malloc((npoints + 1) * sizeof(gdouble));
+
+     npoints = 0;
+     y = ymin;
+     for(ny = 0; ny < surface->ny; ny++)
+       {
+         x = xmin;
+         for(nx = 0; nx < surface->nx; nx++)
+          {
+            gboolean error;
+            fx[npoints] = x;
+            fy[npoints] = y;
+            fz[npoints] = data->function3d(plot, data, x, y, &error);
+
+            x += xstep;
+            npoints++;
+          }
+         y += ystep;
+     }
+
+     old_x = data->x; old_y = data->y; old_z = data->z;
+
+     data->x = fx;
+     data->y = fy;
+     data->z = fz;
+
+     data->num_points = npoints;
+     gtk_plot_surface_real_build_mesh(surface);
+
+     g_free(fx);
+     g_free(fy);
+     g_free(fz);
+  }
+  else if(data->is_iterator){
+     gdouble x, y, z, a, dx, dy, dz, da;
+     gchar *label;
+     gdouble *fx = NULL, *fy = NULL, *fz = NULL;
+     gint iter;
+
+     if(data->iterator_mask & GTK_PLOT_DATA_X)
+       fx = g_new0(gdouble, data->num_points);
+     if(data->iterator_mask & GTK_PLOT_DATA_Y)
+       fy = g_new0(gdouble, data->num_points);
+     if(data->iterator_mask & GTK_PLOT_DATA_Z)
+       fz = g_new0(gdouble, data->num_points);
+
+     for(iter = 0; iter < data->num_points; iter++)
+       {
+          gboolean error;
+          data->iterator (plot, data, iter,
+                          &x, &y, &z, &a, &dx, &dy, &dz, &da, &label, &error);
+
+          if(error)
+            {
+               break;
+            }
+          else
+            {
+              if(data->iterator_mask & GTK_PLOT_DATA_X) fx[iter] = x;
+              if(data->iterator_mask & GTK_PLOT_DATA_Y) fy[iter] = y;
+              if(data->iterator_mask & GTK_PLOT_DATA_Z) fz[iter] = z;
+            }
+       } 
+
+     data->x = fx;
+     data->y = fy;
+     data->z = fz;
+
+     gtk_plot_surface_real_build_mesh(surface);
+
+     g_free(fx);
+     g_free(fy);
+     g_free(fz);
+  }
+  else   
+     gtk_plot_surface_real_build_mesh(surface);
+}
+
+static void
+gtk_plot_surface_real_build_mesh(GtkPlotSurface *surface)
+{
+  GtkPlotData *data;
+  GtkPlot *plot;
+  gint i;
+
+  data = GTK_PLOT_DATA(surface);
+  if(!data->plot) return;
+  plot = data->plot;
+
+  if(data->num_points == 0) return;
+
+  gtk_plot_dt_clear(surface->dt);
+
+  for(i = 0; i < data->num_points; i++){
+    GtkPlotDTnode node;
+    node.x = data->x[i];
+    node.y = data->y[i];
+    node.z = 0.0;
+    if(data->z) node.z = data->z[i];
+    gtk_plot_dt_add_node(surface->dt, node);
+  }
+  gtk_plot_dt_triangulate(surface->dt);
+  gtk_plot_surface_recalc_nodes(surface);
+  surface->recalc_dt = FALSE;
+}
+
+void
+gtk_plot_surface_recalc_nodes(GtkPlotSurface *surface)
+{
+  GtkPlotData *data;
+  GtkPlot *plot;
+  gint i;
+
+  data = GTK_PLOT_DATA(surface);
+  if(!data->plot) return;
+  plot = data->plot;
+
+  for(i= surface->dt->node_0; i < surface->dt->node_cnt; i++){
+    GtkPlotDTnode *node;
+    node = gtk_plot_dt_get_node(surface->dt,i);
+    if(GTK_IS_PLOT3D(plot)){
+      gtk_plot3d_get_pixel(GTK_PLOT3D(plot),
+			   node->x, node->y, node->z,
+			   &node->px, &node->py, &node->pz);
+    } else {
+      gtk_plot_get_pixel(plot,
+			 node->x, node->y,
+			 &node->px, &node->py);
+      node->pz = 0.0;
+    }
+  }
+
+  surface->dt->triangles = g_list_sort(surface->dt->triangles, (GCompareFunc)compare_func);
+}
+
+static gint
+compare_func (gpointer a, gpointer b)
+{
+  GtkPlotDTtriangle *ta, *tb;
+  gdouble cza, czb;
+
+  ta = (GtkPlotDTtriangle *)a;
+  tb = (GtkPlotDTtriangle *)b;
+  cza = MIN(ta->na->pz, ta->nb->pz);
+  cza = MIN(cza, ta->nc->pz);
+  czb = MIN(tb->na->pz , tb->nb->pz);
+  czb = MIN(czb , tb->nc->pz);
+  return(czb > cza ? 1 : (cza == czb ? 0 : -1));
+}
+
+
