@@ -422,6 +422,12 @@ build_example1(GtkWidget *widget)
  sheet=GTK_SHEET(widget);
  colormap = gdk_colormap_get_system();
 
+ gdk_color_parse("light yellow", &color);
+ gdk_color_alloc(gtk_widget_get_colormap(widget), &color);
+ gtk_sheet_set_background(sheet, &color);
+ gdk_color_parse("light blue", &color);
+ gdk_color_alloc(gtk_widget_get_colormap(widget), &color);
+ gtk_sheet_set_grid(sheet, &color);
 
  for(i=0; i<=sheet->maxcol; i++){
    name[0]='A'+i;
@@ -653,6 +659,8 @@ build_example3(GtkWidget *widget)
  
  sheet=GTK_SHEET(widget);
 
+ gtk_sheet_show_grid(sheet, FALSE);
+
  range.row0=0;
  range.rowi=10;
  range.col0=0;
